import json
import os
import re
import sys
import uuid
from datetime import datetime

from rich import box
from rich.color import Color
from rich.console import Console
from rich.pretty import Pretty
from rich.progress import Progress, Column, TextColumn, BarColumn, TimeElapsedColumn, MofNCompleteColumn
from rich.style import Style
from rich.table import Table
from sqlalchemy.exc import NoSuchTableError

from piperider_cli import convert_to_tzlocal, datetime_to_str, clone_directory, \
    raise_exception_when_directory_not_writable
from piperider_cli import event
from piperider_cli.adapter import DbtAdapter
from piperider_cli.assertion_engine import AssertionEngine
from piperider_cli.assertion_engine.recommender import RECOMMENDED_ASSERTION_TAG
from piperider_cli.configuration import Configuration
from piperider_cli.datasource import DataSource
from piperider_cli.exitcode import EC_ERR_TEST_FAILED
from piperider_cli.filesystem import FileSystem
from piperider_cli.profiler import Profiler, ProfilerEventHandler


class RunEventPayload:

    def __init__(self):
        # all fields
        self.tables = 0
        self.columns = []
        self.rows = []
        self.dbt_command = ''
        self.passed_assertions = 0
        self.failed_assertions = 0
        self.passed_dbt_testcases = 0
        self.failed_dbt_testcases = 0
        self.build_in_assertions = 0
        self.custom_assertions = 0
        self.recommended_assertions = 0

    def to_dict(self):
        return self.__dict__


class RichProfilerEventHandler(ProfilerEventHandler):

    def __init__(self, tables):
        table_width, total_width = self._get_width(tables)
        total_column = TextColumn("{task.fields[coft]}", table_column=Column(width=total_width))
        text_column = TextColumn("{task.description}", table_column=Column(width=table_width))
        bar_column = BarColumn(bar_width=80, pulse_style=Style.from_color(Color.from_rgb(244, 164, 96)))
        mofn_column = MofNCompleteColumn(table_column=Column(width=5, justify="right"))
        time_elapsed_column = TimeElapsedColumn()

        self.progress = Progress(total_column, text_column, bar_column, mofn_column, time_elapsed_column)
        self.progress_started = False
        self.tasks = {}
        self.table_total = 0
        self.table_completed = 0

    def _get_width(self, tables):
        return max([len(x) for x in tables]), len(str(len(tables))) * 2 + 2

    def handle_run_start(self, run_result):
        self.progress.start()

    def handle_run_progress(self, run_result, total, completed):
        self.table_total = total

    def handle_run_end(self, run_result):
        self.progress.stop()

    def handle_fetch_metadata_all_start(self):
        print("fetching metadata")

    def handle_fetch_metadata_table_start(self, table_name):
        print(f"fetching metadata for table '{table_name}'")

    def handle_table_start(self, table_result):
        self.table_completed += 1
        table_name = table_result['name']
        padding = ' ' * (len(str(self.table_total)) - len(str(self.table_completed)))
        coft = f'[{padding}{self.table_completed}/{self.table_total}]'
        task_id = self.progress.add_task(table_name, total=None, **dict(coft=coft))
        self.tasks[table_name] = task_id

    def handle_table_progress(self, table_result, total, completed):
        if completed == 0:
            table_name = table_result['name']
            task_id = self.tasks[table_name]
            self.progress.update(task_id, total=total)

    def handle_table_end(self, table_result):
        pass

    def handle_column_start(self, table_name, column_result):
        pass

    def handle_column_end(self, table_name, column_result):
        task_id = self.tasks[table_name]
        self.progress.update(task_id, advance=1)


def _agreed_to_run_recommended_assertions(console: Console, interactive: bool):
    if interactive:
        console.print('Please press enter to continue ...',
                      end=' ')
        confirm = input('').strip().lower()
        return confirm == 'yes' or confirm == 'y' or confirm == ''  # default yes
    else:
        return True


def _agreed_to_generate_recommended_assertions(console: Console, interactive: bool, skip_recommend: bool):
    if skip_recommend:
        return False

    if interactive:
        console.print('Do you want to auto generate recommended assertions for this datasource \[Yes/no]?',
                      end=' ')
        confirm = input('').strip().lower()
        return confirm == 'yes' or confirm == 'y' or confirm == ''  # default yes
    else:
        return True


def _execute_assertions(console: Console, engine, ds_name: str, interaction: bool,
                        output, profiler_result, created_at, skip_recommend: bool):
    # TODO: Implement running test cases based on profiling result
    assertion_engine = AssertionEngine(engine)
    assertion_engine.load_assertions(profiler_result)
    assertion_exist = True if assertion_engine.assertions_content else False

    results = exceptions = []
    if not assertion_exist:
        console.print('[bold yellow]No assertion found[/]')
        if _agreed_to_generate_recommended_assertions(console, interaction, skip_recommend):
            # Generate recommended assertions
            console.rule('Generating Recommended Assertions')
            recommended_assertions = assertion_engine.generate_recommended_assertions(profiler_result)
            for f in recommended_assertions:
                console.print(f'[bold green]Recommended Assertion[/bold green]: {f}')
            if _agreed_to_run_recommended_assertions(console, interaction):
                assertion_engine.load_assertions(profiler_result)
            else:
                console.print(f'[[bold yellow]Skip[/bold yellow]] Executing assertion for datasource [ {ds_name} ]')
                return [], []
        else:
            # Generate assertion templates
            console.rule('Generating Assertion Templates')
            template_assertions = assertion_engine.generate_template_assertions(profiler_result)
            for f in template_assertions:
                console.print(f'[bold green]Template Assertion[/bold green]: {f}')

    # Execute assertions
    if len(assertion_engine.assertions):
        console.rule('Testing')
        results, exceptions = assertion_engine.evaluate_all()
    return results, exceptions


def _show_dbt_test_result(dbt_test_results, title=None, failed_only=False):
    console = Console()
    ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta',
                        box=box.SIMPLE, title=title)
    ascii_table.add_column('Status', style='bold white')
    ascii_table.add_column('Test Subject', style='bold')
    ascii_table.add_column('Assertion', style='bold green')
    ascii_table.add_column('Message', style='bold')

    for r in dbt_test_results:
        if failed_only and r.get('status') == 'passed':
            continue
        success = True if r.get('status') == 'passed' else False
        test_name = r.get('display_name')
        table = r.get('table')
        column = r.get('column')
        target = f'[yellow]{table}[/yellow]'
        if column:
            target = f'{target}.[blue]{column}[/blue]'
        message = r.get('message')

        if success:
            ascii_table.add_row(
                '[[bold green]  OK  [/bold green]]',
                target,
                test_name,
                message
            )
        else:
            ascii_table.add_row(
                '[[bold red]FAILED[/bold red]]',
                target,
                test_name,
                message
            )

    if ascii_table.rows:
        console.print(ascii_table)


def _show_assertion_result(results, exceptions, failed_only=False, single_table=None, title=None):
    console = Console()

    def _wrap_pretty(obj):
        if obj is None:
            return '-'
        return obj if isinstance(obj, str) else Pretty(obj)

    if results:
        ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta',
                            box=box.SIMPLE, title=title)
        ascii_table.add_column('Status', style='bold white')
        ascii_table.add_column('Test Subject', style='bold')
        ascii_table.add_column('Assertion', style='bold green')
        ascii_table.add_column('Expected', style='bold')
        ascii_table.add_column('Actual', style='cyan')

        for assertion in results:
            if single_table and single_table != assertion.table:
                continue
            if failed_only and assertion.result.status():
                continue
            table = assertion.table
            column = assertion.column
            test_function = assertion.result.name
            success = assertion.result.status()
            target = f'[yellow]{table}[/yellow].[blue]{column}[/blue]' if column else f'[yellow]{table}[/yellow]'
            if success:
                ascii_table.add_row(
                    '[[bold green]  OK  [/bold green]]',
                    target,
                    test_function,
                    _wrap_pretty(assertion.result.expected),
                    _wrap_pretty(assertion.result.actual)
                )
            else:
                ascii_table.add_row(
                    '[[bold red]FAILED[/bold red]]',
                    target,
                    test_function,
                    _wrap_pretty(assertion.result.expected),
                    _wrap_pretty(assertion.result.actual)
                )
                if assertion.result.exception:
                    msg = f'[grey11 on white][purple4]{type(assertion.result.exception).__name__}[/purple4](\'{assertion.result.exception}\')[/grey11 on white]'
                    ascii_table.add_row('', '', '', msg)

        if ascii_table.rows:
            console.print(ascii_table)
    # TODO: Handle exceptions
    pass


def _show_recommended_assertion_notice_message(console: Console, results):
    for assertion in results:
        if assertion.result.status() is False and RECOMMENDED_ASSERTION_TAG in assertion.tags:
            console.print(
                f'\n[[bold yellow]Notice[/bold yellow]] You can use command '
                f'"{os.path.basename(sys.argv[0])} generate-assertions" '
                f'to re-generate recommended assertions with new profiling results.')
            break


def _show_dbt_test_result_summary(dbt_test_results, table: str = None):
    if not dbt_test_results:
        return None

    # Prepare DBT Tests Summary
    ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta', box=box.SIMPLE_HEAVY)
    ascii_table.add_column('Table Name', style='bold yellow')
    ascii_table.add_column('#DBT Tests Executed', style='bold blue', justify='right')
    ascii_table.add_column('#DBT Tests Failed', style='bold red', justify='right')

    test_count = {}
    for r in dbt_test_results:
        t = r.get('table')

        if table is not None and t != table:
            continue
        if t not in test_count:
            test_count[t] = dict(total=0, failed=0)

        test_count[t]['total'] += 1
        test_count[t]['failed'] += 1 if r.get('status') != 'passed' else 0

    for t in test_count:
        ascii_table.add_row(
            t,
            Pretty(test_count[t]['total']),
            Pretty(test_count[t]['failed']),
        )

    return ascii_table


def _show_summary(profiled_result, assertion_results, assertion_exceptions, dbt_test_results, table=None):
    console = Console()
    tables = profiled_result.get('tables', []) if table is None else [table]

    # Prepare PipeRider Assertions Summary
    ascii_table = Table(show_header=True, show_edge=True, header_style='bold magenta', box=box.SIMPLE_HEAVY)
    ascii_table.add_column('Table Name', style='bold yellow')
    ascii_table.add_column('#Columns Profiled', style='bold blue', justify='right')
    ascii_table.add_column('#Tests Executed', style='bold blue', justify='right')
    ascii_table.add_column('#Tests Failed', style='bold red', justify='right')

    ascii_dbt_table = _show_dbt_test_result_summary(dbt_test_results)
    for t in tables:
        _show_table_summary(ascii_table, t, profiled_result, assertion_results)

    if ascii_dbt_table:
        # Display DBT Tests Summary
        console.rule('dbt')
        console.print(ascii_dbt_table)
        _show_dbt_test_result(dbt_test_results, failed_only=True, title="Failed DBT Tests")
        if ascii_table.rows:
            console.rule('PipeRider')

    # Display PipeRider Assertions Summary
    if ascii_table.rows:
        console.print(ascii_table)
        _show_assertion_result(assertion_results, assertion_exceptions, failed_only=True,
                               title='Failed Assertions')


def _show_table_summary(ascii_table: Table, table: str, profiled_result, assertion_results):
    profiled_columns = profiled_result['tables'][table].get('col_count')
    num_of_testcases = 0
    num_of_failed_testcases = 0

    if assertion_results:
        for r in assertion_results:
            if r.table == table:
                num_of_testcases += 1
                if not r.result.status():
                    num_of_failed_testcases += 1

    ascii_table.add_row(
        table,
        Pretty(profiled_columns),
        Pretty(num_of_testcases),
        Pretty(num_of_failed_testcases),
    )


def _transform_assertion_result(table: str, results):
    tests = []
    columns = {}
    if results is None:
        return dict(tests=tests, columns=columns)

    for r in results:
        if r.table == table:
            entry = r.to_result_entry()
            if r.column:
                if r.column not in columns:
                    columns[r.column] = []
                columns[r.column].append(entry)
            else:
                tests.append(entry)

    return dict(tests=tests, columns=columns)


def _validate_assertions(console: Console):
    assertion_engine = AssertionEngine(None)
    assertion_engine.load_all_assertions_for_validation()
    results = assertion_engine.validate_assertions()
    # if results
    for result in results:
        # result
        console.print(f'  [[bold red]FAILED[/bold red]] {result.as_user_report()}')

    if results:
        # stop runner
        return True

    # continue to run profiling
    console.print('everything is OK.')
    return False


def _get_table_list(table, default_schema, dbt_adapter):
    tables = None

    if table:
        table = re.sub(f'^({default_schema})\.', '', table)
        tables = [table]

    if dbt_adapter.is_ready():
        dbt_tables = dbt_adapter.list_dbt_tables(default_schema)
        if not dbt_tables:
            return None, "No table found in dbt project."

        if not table:
            tables = dbt_tables
        elif table not in dbt_tables:
            suggestion = ''
            lower_tables = [t.lower() for t in dbt_tables]
            if table.lower() in lower_tables:
                index = lower_tables.index(table.lower())
                suggestion = f"Do you mean '{dbt_tables[index]}'?"
            return None, f"Table '{table}' doesn't exist in dbt project. {suggestion}"

    return tables, None


def prepare_default_output_path(filesystem: FileSystem, created_at, ds):
    latest_symlink_path = os.path.join(filesystem.get_output_dir(), 'latest')
    latest_source = f"{ds.name}-{convert_to_tzlocal(created_at).strftime('%Y%m%d%H%M%S')}"
    output_path = os.path.join(filesystem.get_output_dir(), latest_source)

    if not os.path.exists(output_path):
        os.makedirs(output_path, exist_ok=True)

    # Create a symlink pointing to the latest output directory
    if os.path.islink(latest_symlink_path):
        os.unlink(latest_symlink_path)
    if not os.path.exists(latest_symlink_path):
        os.symlink(latest_source, latest_symlink_path)
    else:
        console = Console()
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return output_path


def _append_descriptions(profile_result):
    for table_v in profile_result['tables'].values():
        table_v['description'] = 'Description: N/A'
        for column_v in table_v['columns'].values():
            column_v['description'] = 'Description: N/A'


def _clean_up_profile_null_properties(table_results):
    removed = []
    for t_metric, t_metric_val in table_results.items():
        if t_metric_val is None:
            removed.append(t_metric)

    for r in removed:
        del table_results[r]

    removed = []
    for col_name, props in table_results.get('columns', {}).items():
        for k, v in props.items():
            if v is None:
                removed.append(dict(col=col_name, key=k))

    for r in removed:
        del table_results['columns'][r['col']][r['key']]


def _append_descriptions_from_assertion(profile_result):
    engine = AssertionEngine(None)
    engine.load_assertion_content()
    for table_name, table_v in engine.assertions_content.items():
        if table_name not in profile_result['tables'] or table_v is None:
            continue
        table_desc = table_v.get('description', '')
        if table_desc:
            profile_result['tables'][table_name]['description'] = f'{table_desc} - via PipeRider'

        columns_content = table_v.get('columns') if table_v.get('columns') else {}
        for column_name, column_v in columns_content.items():
            if column_name not in profile_result['tables'][table_name]['columns'] or column_v is None:
                continue
            column_desc = column_v.get('description', '')
            if column_desc:
                profile_result['tables'][table_name]['columns'][column_name][
                    'description'] = f'{column_desc} - via PipeRider'


def _analyse_and_log_run_event(profiled_result, assertion_results, dbt_test_results, dbt_command):
    tables = profiled_result.get('tables', [])
    event_payload = RunEventPayload()
    event_payload.tables = len(tables)

    # Table info
    for t in tables:
        event_payload.columns.append(profiled_result['tables'][t]['col_count'])
        event_payload.rows.append(profiled_result['tables'][t]['row_count'])

    # Count PipeRider assertions
    for r in assertion_results or []:
        if r.is_builtin:
            event_payload.build_in_assertions += 1
        else:
            event_payload.custom_assertions += 1

        if RECOMMENDED_ASSERTION_TAG in r.tags:
            event_payload.recommended_assertions += 1

        if r.result.status():
            event_payload.passed_assertions += 1
        else:
            event_payload.failed_assertions += 1

    # Count dbt-test cases
    if dbt_test_results:
        for r in dbt_test_results:
            if r.get('status') == 'passed':
                event_payload.passed_dbt_testcases += 1
            else:
                event_payload.failed_dbt_testcases += 1

    event_payload.dbt_command = dbt_command
    event.log_event(event_payload.to_dict(), 'run')


def decorate_with_metadata(profile_result: dict):
    from piperider_cli import __version__
    from piperider_cli.profiler.version import schema_version

    configuration = Configuration.load()
    project_id = configuration.get_telemetry_id()

    profile_result['version'] = __version__
    profile_result['project_id'] = f'{project_id}'
    profile_result['user_id'] = f'{event._collector._user_id}'
    profile_result['metadata_version'] = schema_version()


def _check_test_status(assertion_results, assertion_exceptions, dbt_test_results):
    if assertion_exceptions and len(assertion_exceptions) > 0:
        return False

    if assertion_results:
        for assertion in assertion_results:
            if not assertion.result.status():
                return False

    if dbt_test_results:
        for r in dbt_test_results:
            if r.get('status') == 'failed':
                return False

    return True


class Runner():
    @staticmethod
    def exec(datasource=None, table=None, output=None, interaction=True, skip_report=False, dbt_command='',
             skip_recommend=False, report_dir: str = None):
        console = Console()

        raise_exception_when_directory_not_writable(output)

        configuration = Configuration.load()
        filesystem = FileSystem(report_dir=report_dir)
        datasources = {}
        datasource_names = []
        for ds in configuration.dataSources:
            datasource_names.append(ds.name)
            datasources[ds.name] = ds

        if len(datasource_names) == 0:
            console.print("[bold red]Error: no datasource found[/bold red]")
            return 1

        if datasource and datasource not in datasource_names:
            console.print(f"[bold red]Error: datasource '{datasource}' doesn't exist[/bold red]")
            console.print(f"Available datasources: {', '.join(datasource_names)}")
            return 1

        # Use the first datasource if no datasource is specified
        ds_name = datasource if datasource else datasource_names[0]
        ds = datasources[ds_name]

        passed, reasons = ds.validate()
        if not passed:
            console.print(f"[bold red]Error:[/bold red] The credential of '{ds.name}' is not configured.")
            for reason in reasons:
                console.print(f'    {reason}')
            console.print("[bold yellow]Hint:[/bold yellow]\n  Please execute command 'piperider init' to move forward.")
            return 1

        if not datasource and len(datasource_names) > 1:
            console.print(
                f"[bold yellow]Warning: multiple datasources found ({', '.join(datasource_names)}), using '{ds_name}'[/bold yellow]\n")

        console.print(f'[bold dark_orange]DataSource:[/bold dark_orange] {ds.name}')
        console.rule('Validating')
        err = ds.verify_connector()
        if err:
            console.print(
                f'[[bold red]FAILED[/bold red]] Failed to load the \'{ds.type_name}\' connector.')
            raise err

        try:
            available_tables = ds.verify_connection()
        except Exception as err:
            console.print(
                f'[[bold red]FAILED[/bold red]] Failed to connect the \'{ds.name}\' data source.')
            raise err
        stop_runner = _validate_assertions(console)
        if stop_runner:
            console.print('\n\n[bold red]ERROR:[/bold red] Stop profiling, please fix the syntax errors above.')
            return 1

        default_schema = ds.credential.get('schema')

        dbt_config = ds.args.get('dbt')
        dbt_adapter = DbtAdapter(dbt_config)
        if dbt_config and not dbt_adapter.is_ready():
            raise dbt_adapter.get_error()

        tables = None
        if table:
            tables = [table]

        dbt_test_results = None
        dbt_test_results_compatible = None
        if dbt_command in ['build', 'test'] and dbt_adapter.is_ready():
            dbt_adapter.set_dbt_command(dbt_command)
            dbt_test_results, dbt_test_results_compatible = dbt_adapter.run_dbt_command(table, default_schema)

        console.rule('Profiling')
        run_id = uuid.uuid4().hex
        created_at = datetime.utcnow()
        engine = ds.create_engine()

        if table:
            # cli --table is specified, no inclusion and exclusion applied
            configuration.includes = None
            configuration.excludes = None

        run_result = {}
        profiler = Profiler(engine, RichProfilerEventHandler(tables if tables else available_tables), configuration)
        try:
            profiler_result = profiler.profile(tables)
            run_result.update(profiler_result)
        except NoSuchTableError as e:
            console.print(f"[bold red]Error:[/bold red] No such table '{str(e)}'")
            return 1
        except Exception as e:
            raise Exception(f'Profiler Exception: {type(e).__name__}(\'{e}\')')

        # TODO: refactor input unused arguments
        assertion_results, assertion_exceptions = _execute_assertions(console, engine, ds.name, interaction, output,
                                                                      profiler_result, created_at, skip_recommend)

        run_result['tests'] = []
        if assertion_results or dbt_test_results:
            console.rule('Assertion Results')
            if dbt_test_results:
                console.rule('dbt')
                _show_dbt_test_result(dbt_test_results)
                run_result['tests'].extend(dbt_test_results)
                if assertion_results:
                    console.rule('PipeRider')
            if assertion_results:
                _show_assertion_result(assertion_results, assertion_exceptions)
                run_result['tests'].extend([r.to_result_entry() for r in assertion_results])

        console.rule('Summary')

        if dbt_test_results_compatible:
            for k, v in dbt_test_results_compatible.items():
                if k not in run_result['tables']:
                    continue
                run_result['tables'][k]['dbt_assertion_result'] = v

        for t in run_result['tables']:
            run_result['tables'][t]['piperider_assertion_result'] = _transform_assertion_result(t, assertion_results)
            _clean_up_profile_null_properties(run_result['tables'][t])
        _show_summary(run_result, assertion_results, assertion_exceptions, dbt_test_results)
        _show_recommended_assertion_notice_message(console, assertion_results)

        if dbt_adapter.is_ready():
            dbt_adapter.append_descriptions(run_result, default_schema)
        _append_descriptions_from_assertion(run_result)

        run_result['id'] = run_id
        run_result['created_at'] = datetime_to_str(created_at)
        run_result['datasource'] = dict(name=ds.name, type=ds.type_name)
        decorate_with_metadata(run_result)

        output_path = prepare_default_output_path(filesystem, created_at, ds)
        output_file = os.path.join(output_path, 'run.json')

        with open(output_file, 'w') as f:
            f.write(json.dumps(run_result, separators=(',', ':')))

        if output:
            clone_directory(output_path, output)

        if skip_report:
            console.print(f'Results saved to {output if output else output_path}')

        _analyse_and_log_run_event(run_result, assertion_results, dbt_test_results, dbt_command)

        if not _check_test_status(assertion_results, assertion_exceptions, dbt_test_results):
            return EC_ERR_TEST_FAILED

        return 0



import json
import os

import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TimeElapsedColumn
from ruamel import yaml

from piperider_cli import __version__
from piperider_cli.event import load_user_profile, update_user_profile

PIPERIDER_CLOUD_SERVICE = 'https://cloud.piperider.io/'

SERVICE_ENV_API_KEY = 'PIPERIDER_API_TOKEN'
SERVICE_ENV_SERVICE_KEY = 'PIPERIDER_API_SERVICE'

yml = yaml.YAML()


class CloudServiceHelper:

    def __init__(self):
        self.api_token: str = None
        self.api_service: str = None
        self.user_profile = load_user_profile()
        if self.user_profile is None:
            return
        self.load_configuration()

    @property
    def cloud_host(self):
        if 'localhost' in self.api_service:
            return 'http://localhost:3000'
        return self.api_service

    def load_configuration(self):
        # load from the configuration file first
        # overwrite them if found env vars
        self.api_service = os.environ.get(SERVICE_ENV_SERVICE_KEY,
                                          self.user_profile.get('api_service',
                                                                PIPERIDER_CLOUD_SERVICE))
        self.api_token = os.environ.get(SERVICE_ENV_API_KEY,
                                        self.user_profile.get('api_token'))

    def get_config(self) -> dict:
        config = self.user_profile.get('cloud_config', {})
        return config if config else {}

    def update_config(self, options: dict):
        cloud_config = self.get_config()
        cloud_config.update(options)
        update_user_profile({'cloud_config': cloud_config})

    def update_api_token(self):
        if self.api_token:
            update_user_profile({'api_token': self.api_token})
        else:
            update_user_profile({'api_token': None})

    def url(self, uri_path: str):
        if uri_path and not uri_path.startswith('/'):
            uri_path = f'/{uri_path}'

        if self.api_service and self.api_service.endswith('/'):
            return f'{self.api_service[:-1]}{uri_path}'
        return f'{self.api_service}{uri_path}'

    def auth_headers(self):
        return {
            'User-Agent': f'PipeRider CLI/{__version__}',
            'Authorization': f'Bearer {self.api_token}',
        }

    def http_get(self, uri_path):
        try:
            return requests.get(
                self.url(uri_path),
                headers=self.auth_headers()) \
                .json()
        except BaseException:
            return None

    def validate(self) -> (bool, dict):
        if not (self.api_token and self.api_service):
            return False, None

        json_resp = self.http_get('/api/users/me')
        if json_resp and 'email' in json_resp:
            return True, json_resp
        return False, None


class PipeRiderCloud:

    def __init__(self):
        self.service = CloudServiceHelper()
        self.config: dict = self.service.get_config()
        try:
            self.available, self.me = self.service.validate()
        except BaseException:
            self.available = False
            self.me = None

    def update_config(self, options: dict):
        self.service.update_config(options)
        self.config = self.service.get_config()

    def validate(self, api_token=None):
        if api_token:
            self.service.api_token = api_token

        try:
            self.available, self.me = self.service.validate()
        except BaseException:
            self.available = False
            self.me = None

        return self.available

    def logout(self):
        self.service.api_token = None
        self.service.update_api_token()

    def magic_login(self, email):
        if self.available:
            return True
        login_url = self.service.url('/api/session/new')
        response = requests.post(
            login_url,
            headers={'Content-type': 'application/json', 'Accept': 'text/plain'},
            data=json.dumps({'email': email})
        )
        if response.status_code == 200:
            return response.json()
        return None

    def upload_report(self, file_path, show_progress=True):
        # TODO validate project name
        if not self.available:
            self.raise_error()

        upload_progress = None
        task_id = None

        def _upload_callback(monitor):
            if show_progress and upload_progress:
                upload_progress.update(task_id, completed=monitor.bytes_read)

        with open(file_path, 'rb') as file:
            encoder = MultipartEncoder(
                fields={'file': ('run.json', file)},
            )
            m = MultipartEncoderMonitor(encoder, _upload_callback)

            url = self.service.url('/api/reports/upload')
            headers = self.service.auth_headers()
            headers['Content-Type'] = m.content_type

            if show_progress:
                upload_progress = Progress(
                    TextColumn("[bold magenta]{task.description}[/bold magenta]"),
                    BarColumn(bar_width=80),
                    DownloadColumn(binary_units=True),
                    TimeElapsedColumn(),
                )
                task_id = upload_progress.add_task(description=file_path, total=encoder.len)
                upload_progress.start()

            response = requests.post(url, data=m, headers=headers)

            if show_progress:
                upload_progress.stop()

            return response.json()

    def raise_error(self):
        raise ValueError("Service not available or configuration invalid")

    def has_configured(self):
        return self.service.api_token is not None


if __name__ == '__main__':
    cloud = PipeRiderCloud()
    cloud.upload_report(json.dumps(dict(foo='barbar')))



import os
import webbrowser
from typing import List

import inquirer
from rich import box
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from piperider_cli import datetime_to_str, str_to_datetime
from piperider_cli.cloud import PipeRiderCloud
from piperider_cli.compare_report import CompareReport, RunOutput
from piperider_cli.datasource import FANCY_USER_INPUT
from piperider_cli.filesystem import FileSystem

console = Console()
piperider_cloud = PipeRiderCloud()


def ask_login_info() -> str:
    console.print('Please provide available email account to login')
    if FANCY_USER_INPUT:
        account = inquirer.text('Email address', validate=lambda _, x: '@' in x)
    else:
        while True:
            account = Prompt.ask('[[yellow]?[/yellow]] Email address')
            if '@' in account:
                break

    response = piperider_cloud.magic_login(account)
    if response is None or response.get('success') is False:
        console.print('[[red]Error[/red]] Login failed. Please try again.')
        return None

    if response.get('link'):
        webbrowser.open(response.get('link'))

    console.print('Please paste the api token from magic link. The link has been sent to your email address.')
    while True:
        api_token = Prompt.ask('[[yellow]?[/yellow]] API token')
        if len(api_token) > 0:
            break
    return api_token


def verify_api_token(api_token) -> bool:
    if piperider_cloud.validate(api_token) is False:
        console.print('[[red]Error[/red]] Invalid API Token. Please try again.')
        return False

    # Write API Token back to user profile
    piperider_cloud.service.update_api_token()
    return True


def location_mapper(loc_id):
    map = {
        'us': 'North America',
        'eu': 'Europe',
        'ap': 'Asia Pacific'
    }
    return map.get(loc_id, 'N/A')


def show_user_info():
    user = piperider_cloud.me
    if user is None:
        return
    ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                        box=box.SIMPLE, title='User Profile')
    ascii_table.add_column('Email', justify='left', style='cyan')
    ascii_table.add_column('Username', justify='left')
    ascii_table.add_column('Full Name', justify='left')
    ascii_table.add_column('Storage Location', justify='left')
    ascii_table.add_column('Timezone', justify='left')
    ascii_table.add_row(
        user.get('email'),
        user.get('username', 'N/A'),
        user.get('fullname', 'N/A'),
        location_mapper(user.get('data_storage_location')),
        user.get('timezone_id')
    )
    console.print(ascii_table)
    pass


def check_default_config(options: dict):
    if options is None:
        return

    cloud_config = piperider_cloud.config

    auto_upload_flag = options.get('auto_upload')

    if cloud_config.get('auto_upload') is not None and auto_upload_flag is None:
        # Skip if auto_upload is already set
        return
    if auto_upload_flag is None and cloud_config.get('auto_upload') is None:
        console.print('Please select default behavior for auto upload')
        if FANCY_USER_INPUT:
            auto_upload_flag = inquirer.confirm('Auto upload reports to cloud', default=True)
        else:
            auto_upload_flag = Prompt.ask('[[yellow]?[/yellow]] Auto upload reports to cloud', choices=['y', 'n'],
                                          default='y') == 'y'
    console.print(f'[[bold green]Config[/bold green]] Default auto upload behavior is set to {auto_upload_flag}')
    piperider_cloud.update_config({'auto_upload': auto_upload_flag})


def select_reports(report_dir=None, datasource=None) -> List[RunOutput]:
    filesystem = FileSystem(report_dir=report_dir)
    selector = CompareReport(filesystem.get_output_dir(), None, None, datasource=datasource)
    console.rule('Select Reports to Upload')
    return selector.select_multiple_reports(action='upload')


def upload_to_cloud(report: RunOutput, debug=False) -> dict:
    response = piperider_cloud.upload_report(report.path)
    # TODO refine the output when API is ready

    if response.get('success') is True:
        project_id = response.get('project_id')
        report_id = response.get('id')
        if project_id and report_id:
            report_url = f'{piperider_cloud.service.cloud_host}/projects/{project_id}/reports/{report_id}'
        else:
            report_url = 'N/A'
        return {
            'success': True,
            'message': response.get("message"),
            'report_url': report_url,
            'file_path': report.path
        }

    if debug:
        console.print(response)
    return {
        'success': False,
        'message': response.get("message"),
        'report_url': 'N/A',
        'file_path': report.path
    }


class CloudConnector():
    @staticmethod
    def is_login() -> bool:
        return piperider_cloud.available

    @staticmethod
    def is_auto_upload() -> bool:
        return piperider_cloud.config.get('auto_upload', False)

    @staticmethod
    def config_auto_upload(flag: bool):
        console.print(f'[[bold green]Config[/bold green]] Default auto upload behavior is set to {flag}')
        piperider_cloud.update_config({'auto_upload': flag})

    @staticmethod
    def login(api_token=None, options: dict = None, debug=False):

        if piperider_cloud.available:
            console.rule('Already Logged In')
            show_user_info()
            check_default_config(options)
            return 0

        if api_token is None:
            api_token = ask_login_info()

        if api_token:
            if verify_api_token(api_token):
                console.rule('Login Successful')
                show_user_info()
                check_default_config(options)
                return 0

        console.rule('Login Failed', style='red')
        return 1

    @staticmethod
    def logout():

        if piperider_cloud.available is False:
            console.rule('Already Logout', style='yellow')
            return 0

        piperider_cloud.logout()
        console.rule('Logout')
        return 0

    @staticmethod
    def upload_latest_report(report_dir=None, debug=False) -> int:
        filesystem = FileSystem(report_dir=report_dir)
        latest_report_path = os.path.join(filesystem.get_output_dir(), 'latest', 'run.json')
        return CloudConnector.upload_report(latest_report_path, debug=debug)

    @staticmethod
    def upload_report(report_path=None, report_dir=None, datasource=None, debug=False) -> int:
        if piperider_cloud.available is False:
            console.rule('Please login PipeRider Cloud first', style='red')
            return 1

        rc = 0
        results = []
        if report_path is None:
            reports = select_reports(report_dir=report_dir, datasource=datasource)
            console.rule('Uploading Reports')
            for r in reports:
                response = upload_to_cloud(r, debug)
                if response.get('success') is False:
                    rc = 1
                response['name'] = r.name
                response['created_at'] = r.created_at
                results.append(response)
        else:
            console.rule('Uploading Report')
            report = RunOutput(report_path)
            response = upload_to_cloud(report, debug)
            if response.get('success') is False:
                rc = 1
            response['name'] = report.name
            response['created_at'] = report.created_at
            results.append(response)

        console.rule('Upload Completed')
        ascii_table = Table(show_header=True, show_edge=True, header_style="bold magenta",
                            box=box.SIMPLE)
        ascii_table.add_column('Status', justify='left', style='cyan')
        ascii_table.add_column('Name', justify='left')
        ascii_table.add_column('Created At', justify='left')
        ascii_table.add_column('Report URL', justify='left', no_wrap=True)
        ascii_table.add_column('Message', justify='left')

        for response in results:
            status = '[bold green]Success[/bold green]' if response.get(
                'success') else '[bold yellow]Skipped[/bold yellow]'
            url = f"[deep_sky_blue1]{response.get('report_url', 'N/A')}[/deep_sky_blue1]"
            message = response.get('message')
            created_at = datetime_to_str(str_to_datetime(response.get('created_at')), to_tzlocal=True)
            ascii_table.add_row(status, response.get('name'), created_at, url, message)

        console.print(ascii_table)
        return rc



import io
import json
import os
import sys
from datetime import datetime

import inquirer
import readchar
from rich.console import Console

import piperider_cli.hack.inquirer as inquirer_hack
from piperider_cli import datetime_to_str, str_to_datetime, clone_directory, \
    raise_exception_when_directory_not_writable
from piperider_cli.filesystem import FileSystem
from piperider_cli.generate_report import setup_report_variables


class RunOutput(object):
    def __init__(self, path):
        self.path = path
        self.name = None
        self.created_at = None

        self.table_count = 0
        self.pass_count = 0
        self.fail_count = 0

        try:
            with open(path, 'r') as f:
                run_result = json.load(f)
                self.name = run_result['datasource']['name']
                self.created_at = run_result['created_at']

                tables = run_result.get('tables', {})
                self.table_count = len(tables.keys())

                for test in run_result.get('tests', []):
                    if test.get('status') == 'passed':
                        self.pass_count += 1
                    else:
                        self.fail_count += 1
        except Exception as e:
            if isinstance(e, json.decoder.JSONDecodeError):
                raise json.decoder.JSONDecodeError(
                    f'Invalid JSON in file "{path}"', e.doc, e.pos)
            raise e

    def verify(self) -> bool:
        # TODO: add some verification logic
        return True

    def load(self):
        with open(self.path, 'r') as f:
            data = json.load(f)
        return data

    def __str__(self):
        created_at_str = datetime_to_str(str_to_datetime(self.created_at),
                                         to_tzlocal=True)

        return f'{self.name:12} ' \
               f'#table={self.table_count:<6} ' \
               f'#pass={self.pass_count:<5} ' \
               f'#fail={self.fail_count:<5} ' \
               f'{created_at_str}'


def join(base, target):
    '''
    Join base and target to a dict which

    keys = (base keys) +  (target keys)
    result[keys] = {base: {...}, target: {...}

    :param base:
    :param target:
    :return:
    '''
    if not base:
        base = dict()
    if not target:
        target = dict()
    result = dict()

    joined = target.copy()
    joined.update(base)

    for key in joined.keys():
        value = dict()
        value['base'] = base.get(key)
        value['target'] = target.get(key)
        result[key] = value
    return result


def value_with_annotation(key, annotation=None):
    annotation_str = f" ({annotation})" if annotation else ''
    return f"{key}{annotation_str}"


def value_with_change(base, target):
    if target is None:
        return ''
    elif base is None:
        return target
    elif base != target:
        return f"~~{base}~~<br/>{target}"
    else:
        return target


def value_with_delta(base, target, percentage=False):
    if target is None:
        return ''

    delta = ''
    if base is not None and target is not None:
        delta = f" ({'+' if target >= base else ''}{target - base})"
    return f"{target}{delta}"


class ComparisonData(object):
    def __init__(self, base, target):
        self._id = datetime.now().strftime("%Y%m%d%H%M%S")
        self._base = base
        self._target = target

    def id(self):
        return self._id

    def to_json(self):
        output = dict(
            created_at=datetime_to_str(datetime.utcnow()),
            base=self._base,
            # TODO: rename input -> target in schema and result json
            input=self._target,
        )
        return json.dumps(output, separators=(',', ':'))

    def to_summary_markdown(self):
        base = self._base.get('tables')
        target = self._target.get('tables')

        out = io.StringIO()

        # Comparison summary
        out.write(self._render_compare_markdown(base, target))

        # Per-table summary
        out.write("<details>\n")
        out.write("<summary>Tables Summary</summary>\n")
        out.write("<blockquote>\n")
        out.write("\n")
        joined = join(base, target)
        for table_name in joined.keys():
            joined_table = joined[table_name]

            columns_b = joined_table.get('base').get('columns') if joined_table.get('base') else None
            columns_t = joined_table.get('target').get('columns') if joined_table.get('target') else None

            out.write(self._render_table_summary_markdown(table_name, columns_b, columns_t))
        out.write("</blockquote></details>")
        return out.getvalue()

    def _render_compare_markdown(self, base, target):
        out = io.StringIO()
        joined = join(base, target)
        print("Table | Rows | Columns ", file=out)
        print("--- | --- | ---", file=out)
        for table_name in joined.keys():
            joined_table = joined[table_name]
            b = joined_table.get('base')
            t = joined_table.get('target')

            annotation = None
            if b is None:
                annotation = '+'
            elif t is None:
                annotation = '-'

            rows_b, cols_b = (b.get('row_count', 0), b.get('col_count', 0)) if b else (None, None)
            rows_t, cols_t = (t.get('row_count', 0), t.get('col_count', 0)) if t else (None, None)

            out.write(f"{value_with_annotation(table_name, annotation)} | ")
            out.write(f"{value_with_delta(rows_b, rows_t)} | ")
            out.write(f"{value_with_delta(cols_b, cols_t)} \n")

        return f"""<details>
<summary>Comparison Summary</summary>

{out.getvalue()}
</details>
"""

    def _render_table_summary_markdown(self, table_name, base, target):
        out = io.StringIO()
        joined = join(base, target)
        print("Column | Type | Count", file=out)
        print("--- | --- | --- ", file=out)
        for column_name in joined.keys():
            joined_column = joined[column_name]
            b = joined_column.get('base')
            t = joined_column.get('target')

            schema_type_b, count_b = (b.get('schema_type', ''), b.get('samples', 0)) if b else (None, None)
            schema_type_t, count_t = (t.get('schema_type', ''), t.get('samples', 0)) if t else (None, None)

            annotation = None
            if b is None:
                annotation = '+'
            elif t is None:
                annotation = '-'
            elif schema_type_b != schema_type_t:
                annotation = '!'

            out.write(f"{value_with_annotation(column_name, annotation)} | ")
            out.write(f"{value_with_change(schema_type_b, schema_type_t)} | ")
            out.write(f"{value_with_delta(count_b, count_t)} \n")

        return f"""<details>
<summary>{table_name}</summary>

{out.getvalue()}
</details>
"""


def prepare_default_output_path(filesystem: FileSystem, created_at):
    latest_symlink_path = os.path.join(filesystem.get_comparison_dir(), 'latest')
    latest_source = created_at
    comparison_path = os.path.join(filesystem.get_comparison_dir(), created_at)

    if not os.path.exists(comparison_path):
        os.makedirs(comparison_path, exist_ok=True)

    # Create a symlink pointing to the latest comparison directory
    if os.path.islink(latest_symlink_path):
        os.unlink(latest_symlink_path)
    if not os.path.exists(latest_symlink_path):
        os.symlink(latest_source, latest_symlink_path)
    else:
        console = Console()
        console.print(f'[bold yellow]Warning: {latest_symlink_path} already exists[/bold yellow]')

    return comparison_path


class CompareReport(object):
    def __init__(self, profiler_output_path, a=None, b=None, datasource=None):
        self.profiler_output_path = profiler_output_path
        self.console = Console()
        self.a: RunOutput = RunOutput(a) if a else None
        self.b: RunOutput = RunOutput(b) if b else None
        self.datasource = datasource

    def list_existing_outputs(self, output_search_path=None):
        """
        List existing profiler outputs.
        """

        def _walk_throw_runs(path):
            outputs = []
            for root, dirs, _ in os.walk(path):
                for dir in dirs:
                    if dir == 'latest':
                        continue
                    run_json = os.path.join(root, dir, 'run.json')
                    if not os.path.exists(run_json):
                        continue
                    output = RunOutput(run_json)
                    if self.datasource and output.name != self.datasource:
                        continue
                    outputs.append(output)
            outputs.sort(key=lambda x: (x.name, x.created_at), reverse=True)
            return outputs

        if output_search_path is None:
            output_search_path = self.profiler_output_path

        return _walk_throw_runs(output_search_path)

    def get_the_last_two_reports(self):
        outputs = self.list_existing_outputs()
        outputs.sort(key=lambda x: x.created_at)
        outputs = list(filter(lambda x: x.name == outputs[-1].name, outputs))
        if len(outputs) < 2:
            return None, None
        return outputs[-2:]

    def select_reports(self, use_last_two=None):
        if use_last_two:
            self.a, self.b = self.get_the_last_two_reports()
        elif self.a is None and self.b is None:
            self.a, self.b = self.select_two_reports()
        elif self.a and self.b is None:
            self.b = self.select_one_report()
            pass
        elif self.a is None and self.b:
            self.a = self.select_one_report()

        if self.a and self.b:
            self.console.print('Selected reports:')
            self.console.print(f'  Base:   {self.a.path}')
            self.console.print(f'  Target: {self.b.path}')
            return True
        return False

    def select_multiple_reports(self, action='compare', limit=None):
        def _report_validater(answers, current) -> bool:
            if limit is None:
                return len(current) > 0
            else:
                return len(current) == limit

        profiler_outputs = self.list_existing_outputs()
        arrow_alias_msg = ''
        if sys.platform == "win32" or sys.platform == "cygwin":
            # change readchar key UP & DOWN by 'w' and 's'
            readchar.key.UP = 'w'
            readchar.key.DOWN = 's'
            arrow_alias_msg = " 'w' to Up, 's' to Down,"

        if limit is None:
            if len(profiler_outputs) == 0:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            questions = [
                inquirer.Checkbox('profiler_output',
                                  message=f"Please select the reports to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                  choices=profiler_outputs,
                                  carousel=True,
                                  validate=_report_validater,
                                  )
            ]
        else:
            if len(profiler_outputs) < limit:
                raise Exception("Not enough reports to compare. Please run 'piperider run' first.")
            report_msg = 'a report' if limit == 1 else f'the {limit} reports'
            questions = [
                inquirer_hack.LimitedCheckboxQuestion('profiler_output',
                                                      message=f"Please select {report_msg} to {action} ({arrow_alias_msg} SPACE to select, and ENTER to confirm )",
                                                      choices=profiler_outputs,
                                                      carousel=True,
                                                      validate=_report_validater,
                                                      limited=limit,
                                                      )
            ]

        answers = inquirer_hack.prompt_ex(questions, raise_keyboard_interrupt=True)
        if answers:
            return answers['profiler_output']
        return None

    def select_one_report(self, action='compare') -> RunOutput:
        answers = self.select_multiple_reports(action=action, limit=1)
        if answers:
            return answers[0]
        else:
            return None

    def select_two_reports(self, action='compare') -> (RunOutput, RunOutput):
        """
        Select multiple files from a list of files.
        """

        answers = self.select_multiple_reports(action=action, limit=2)
        if answers:
            return answers[0], answers[1]
        return None, None

    def generate_data(self) -> ComparisonData:
        if self.a is None or self.b is None:
            raise Exception("Please select reports to compare first.")

        return ComparisonData(self.a.load(), self.b.load())

    @staticmethod
    def exec(*, a=None, b=None, last=None, datasource=None, report_dir=None, output=None, debug=False):
        console = Console()

        filesystem = FileSystem(report_dir=report_dir)
        raise_exception_when_directory_not_writable(output)

        report = CompareReport(filesystem.get_output_dir(), a, b, datasource=datasource)
        if not report.select_reports(use_last_two=last):
            raise Exception('No valid reports found')
        comparison_data = report.generate_data()

        from piperider_cli import data
        report_template_dir = os.path.join(os.path.dirname(data.__file__), 'report', 'comparison-report')
        with open(os.path.join(report_template_dir, 'index.html')) as f:
            report_template_html = f.read()

        def output_report(directory):
            clone_directory(report_template_dir, directory)
            filename = os.path.join(directory, 'index.html')
            with open(filename, 'w') as f:
                html = setup_report_variables(report_template_html, False, comparison_data.to_json())
                f.write(html)

        def output_summary(directory):
            filename = os.path.join(directory, 'summary.md')
            with open(filename, 'w') as f:
                f.write(comparison_data.to_summary_markdown())

        data_id = comparison_data.id()
        default_report_directory = prepare_default_output_path(filesystem, data_id)
        output_report(default_report_directory)
        output_summary(default_report_directory)
        report_path = os.path.join(filesystem.get_comparison_dir(), 'latest', 'index.html')

        if output:
            clone_directory(default_report_directory, output)
            report_path = os.path.join(output, 'index.html')

        console.print()
        console.print(f"Comparison report: {report_path}")

        if debug:
            # Write comparison data to file
            with open(os.path.join(default_report_directory, 'comparison_data.json'), 'w') as f:
                f.write(comparison_data.to_json())



from piperider_cli.error import PipeRiderConnectorError
from . import DataSource
from .field import TextField, PasswordField, NumberField


class PostgresDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'postgres', **kwargs)
        self.fields = [
            TextField('host', description='Host URL'),
            NumberField('port', default=5432, description='Port'),
            TextField('user', description='Username'),
            PasswordField('password', description='Password'),
            TextField('dbname', description='Database'),
            TextField('schema', default='public', description='Schema', optional=True),
        ]

    def validate(self):
        if self.type_name != 'postgres':
            raise ValueError('type name should be snowflake')
        return self._validate_required_fields()

    def to_database_url(self):
        credential = self.credential
        host = credential.get('host')
        port = credential.get('port')
        user = credential.get('user')
        password = credential.get('password')
        dbname = credential.get('dbname')
        return f"postgresql+psycopg2://{user}:{password}@{host}:{port}/{dbname}"

    def engine_args(self):
        return dict(connect_args={'connect_timeout': 5})

    def verify_connector(self):
        try:
            import psycopg2
            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'postgres')



from abc import ABCMeta, abstractmethod
from typing import List

from rich.console import Console
from rich.markup import escape

from piperider_cli.adapter import DbtAdapter
from piperider_cli.assertion_engine import AssertionEngine, ValidationResult
from piperider_cli.cloud import PipeRiderCloud
from piperider_cli.configuration import Configuration, PIPERIDER_CONFIG_PATH
from piperider_cli.error import DbtAdapterCommandNotFoundError, DbtCommandNotFoundError, PipeRiderError

CONSOLE_MSG_PASS = '[bold green]✅ PASS[/bold green]\n'
CONSOLE_MSG_FAIL = '[bold red]😱 FAILED[/bold red]\n'
CONSOLE_MSG_ALL_SET = '[bold]🎉 You are all set![/bold]\n'

piperider_cloud = PipeRiderCloud()


class AbstractChecker(metaclass=ABCMeta):
    console = Console()

    @abstractmethod
    def check_function(self, configurator: Configuration) -> (bool, str):
        pass


class CheckingHandler(object):
    def __init__(self):
        self.configurator = None
        self.checker_chain = []
        self.console = Console()

    def set_checker(self, name: str, checker: AbstractChecker):
        self.checker_chain.append({'name': name, 'cls': checker()})

    def execute(self):
        if not self.configurator:
            try:
                self.configurator = Configuration.load()
            except Exception:
                pass

        for checker in self.checker_chain:
            self.console.print(f'Check {checker["name"]}:')
            passed, error_msg = checker['cls'].check_function(self.configurator)
            if not passed:
                hint = None
                if isinstance(error_msg, list):
                    error_msg = ', '.join(str(e) for e in error_msg)
                elif isinstance(error_msg, PipeRiderError):
                    hint = error_msg.hint
                self.console.print(CONSOLE_MSG_FAIL)
                self.console.print(f"[bold red]Error:[/bold red] {checker['cls'].__class__.__name__}: {error_msg}")
                if hint:
                    self.console.print(f'[bold yellow]Hint[/bold yellow]:\n  {escape(hint)}')
                return False
            self.console.print(CONSOLE_MSG_PASS)

        self.console.print(CONSOLE_MSG_ALL_SET)
        return True


class CheckConfiguration(AbstractChecker):
    def check_function(self, configurator: Configuration) -> (bool, str):
        if not configurator:
            self.console.print(f'  {PIPERIDER_CONFIG_PATH}: [[bold red]FAILED[/bold red]]')
            return False, 'No configuration found'
        self.console.print(f'  {PIPERIDER_CONFIG_PATH}: [[bold green]OK[/bold green]]')
        return True, ''


class CheckDataSources(AbstractChecker):
    def check_function(self, configurator: Configuration) -> (bool, str):
        all_passed = True
        failed_reasons = []
        for ds in configurator.dataSources:
            passed, reasons = ds.validate()
            if passed:
                self.console.print(f'  {ds.name}: [[bold green]OK[/bold green]]')
            else:
                all_passed = False
                self.console.print(f'  {ds.name}: [[bold red]FAILED[/bold red]]')
                for reason in reasons:
                    self.console.print(f'    {reason}')
                failed_reasons.extend(reasons)
        return all_passed, failed_reasons


class CheckConnections(AbstractChecker):
    def check_function(self, configurator: Configuration) -> (bool, str):
        all_passed = True
        reason = ''
        for ds in configurator.dataSources:
            dbt = ds.args.get('dbt')
            name = ds.name
            type = ds.type_name

            if dbt:  # dbt provider
                self.console.print(f'  DBT: {ds.type_name} > {dbt["profile"]} > {dbt["target"]} ', end='')
                err = DbtAdapter.check_dbt_command(dbt)
                if err and isinstance(err, DbtCommandNotFoundError):
                    self.console.print(f'[[bold red]FAILED[/bold red]] reason: {err}')
                    all_passed = False
                    reason = err
                    self.console.print(f"\n{DbtAdapterCommandNotFoundError(type).hint}\n")
                else:
                    self.console.print('[[bold green]OK[/bold green]]')

            self.console.print(f'  Name: {name}')
            self.console.print(f'  Type: {type}')

            err = ds.verify_connector()
            if err:
                all_passed = False
                reason = err
                self.console.print(f'  connector: [[bold red]FAILED[/bold red]] reason: {err}')
                self.console.print(f'\n{escape(err.hint)}\n')
                continue
            else:
                self.console.print('  connector: [[bold green]OK[/bold green]]')

            try:
                available_tables = ds.verify_connection()
                self.console.print(f'  Available Tables: {available_tables}')
                self.console.print('  Connection: [[bold green]OK[/bold green]]')
            except Exception as e:
                self.console.print(f'  Connection: [[bold red]FAILED[/bold red]] reason: {e}')
                all_passed = False
                reason = e

        return all_passed, reason


class CheckAssertionFiles(AbstractChecker):
    def check_function(self, configurator: Configuration) -> (bool, str):
        engine = AssertionEngine(None)
        passed_files, failed_files = engine.load_all_assertions_for_validation()
        results: List[ValidationResult] = engine.validate_assertions()

        for file in passed_files:
            self.console.print(f'  {file}: [[bold green]OK[/bold green]]')

        for file in failed_files:
            self.console.print(f'  {file}: [[bold red]FAILED[/bold red]]')

        newline_section = False
        validate_fail = False
        error_msg = ''
        for result in results:
            if result.has_errors():
                if not newline_section:
                    self.console.line()
                    newline_section = True
                self.console.print(f'  [[bold red]FAILED[/bold red]] {result.as_user_report()}')
                validate_fail = True

        if validate_fail or len(failed_files):
            error_msg = 'Syntax problem of PipeRider assertion yaml files'
            self.console.line()

        return error_msg == '', error_msg


class CloudAccountChecker(AbstractChecker):
    def check_function(self, configurator: Configuration) -> (bool, str):
        if not piperider_cloud.available:
            msg = 'API Token is configured, but cannot connect to the service.'
            self.console.print(f'  [bold red]{msg}[/bold red]')
            return False, msg
        else:
            self.console.print(f"  Run as user: [bold green]{piperider_cloud.me.get('email')}[/bold green]")
            self.console.print(f"    User Name: [bold green]{piperider_cloud.me.get('username', 'N/A')}[/bold green]")
            self.console.print(f"    Full Name: [bold green]{piperider_cloud.me.get('fullname', 'N/A')}[/bold green]")
            self.console.print(f"  Auto Upload: {piperider_cloud.config.get('auto_upload', False)}")
            return True, ""


class Validator():
    @staticmethod
    def diagnose():
        handler = CheckingHandler()
        handler.set_checker('config files', CheckConfiguration)
        handler.set_checker('format of data sources', CheckDataSources)
        handler.set_checker('connections', CheckConnections)
        handler.set_checker('assertion files', CheckAssertionFiles)
        if piperider_cloud.has_configured():
            handler.set_checker('cloud account', CloudAccountChecker)
        return handler.execute()



import os

from rich.console import Console

from piperider_cli.configuration import PIPERIDER_WORKSPACE_NAME

piperider_config = os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME, 'config.yml')
report_directory = os.path.join(os.getcwd(), PIPERIDER_WORKSPACE_NAME, 'outputs')


def piperider_initialized() -> bool:
    return os.path.exists(piperider_config)


def number_of_reports():
    if not os.path.exists(report_directory):
        return 0
    return len([x for x in os.listdir(report_directory) if x != 'latest'])


class Guide(object):
    """
    # conditional suggestions

    no .piperider ==> init
    init ==> diagnose
    after diagnose if reports <2 ==> run
    after run if only one report ==> run for compare-report
    after run if ==2 reports ==> compare-report
    """

    def __init__(self):
        self.console = Console()

    def show_tips(self, command_name: str):

        if command_name == 'version':
            return

        if command_name != 'init' and not piperider_initialized():
            self.show("Piperider is not initialized. Please execute command 'piperider init' to move forward.")
            return

        if command_name == 'init':
            self.show("Please execute command 'piperider diagnose' to verify configuration")
            return

        if command_name == 'diagnose' and number_of_reports() < 1:
            self.show("Please execute command 'piperider run' to generate your first report")
            return

        if command_name == 'diagnose' and number_of_reports() < 2:
            self.show("Please execute command 'piperider run' to generate your second report")
            return

        if command_name == 'run' and number_of_reports() == 1:
            self.show("Please execute command 'piperider run' to generate your second report")
            return

        if command_name == 'run' and number_of_reports() == 2:
            self.show("Please execute command 'piperider compare-reports' to get the comparison report")
            return

        if command_name == 'generate-report' and number_of_reports() == 1:
            self.show("Please execute command 'piperider run' to generate your second report")
            return

        if command_name == 'generate-report' and number_of_reports() == 2:
            self.show("Please execute command 'piperider compare-reports' to get the comparison report")
            return

    def show(self, description):
        console = self.console
        console.line()
        console.print("Next step:")
        console.print(f'  {description}')
        console.line()



import os
import sys

import inquirer
import requests
from rich.console import Console
from rich.prompt import Prompt

from piperider_cli import __version__, get_user_id

USER_FEEDBACK_API_TOKEN = '371aa67a-a6eb-44cc-9d02-4a05810f24ac'


def _feedback_collector(message, email=None, version=__version__):
    try:
        user_id = get_user_id()
        json_payload = {
            'user_id': user_id,
            'message': message,
            'email': email,
            'version': version,
        }
        url = f'https://api.aws.piperider.io/v1/feedback?token={USER_FEEDBACK_API_TOKEN}'
        requests.post(url, json=json_payload)
    except Exception:
        pass


def _should_use_fancy_user_input() -> bool:
    env_flag = os.environ.get('PIPERIDER_FANCY_USER_INPUT', 'true').lower() == 'true'
    is_a_tty = sys.stdin.isatty() and sys.stdout.isatty()
    if env_flag is False:
        return False
    if is_a_tty is False:
        return False
    if sys.platform == "darwin" or sys.platform == "linux":
        return True
    else:
        return False


FANCY_USER_INPUT = _should_use_fancy_user_input()


def _feedback_by_inquirer():
    console = Console()
    console.print('[bold dark_orange]Please specify your feedback messages[/bold dark_orange]')
    questions = [
        inquirer.Editor('message', message="Feedback"),
        inquirer.Text('email', message="Contact email (Optional)")
    ]
    feedback = inquirer.prompt(questions, raise_keyboard_interrupt=True)
    _feedback_collector(feedback['message'], feedback['email'])


def _feedback_by_rich():
    console = Console()
    console.print('[bold dark_orange]Please specify your feedback messages[/bold dark_orange]')
    message = Prompt.ask("Feedback")
    email = Prompt.ask("Contact email (Optional)")
    _feedback_collector(message, email)


class Feedback:
    @staticmethod
    def exec():
        if FANCY_USER_INPUT:
            _feedback_by_inquirer()
        else:
            _feedback_by_rich()

    @staticmethod
    def suggest_datasource(data_source_name):
        _feedback_collector(f'Feature request datasource: {data_source_name}')
        pass



from datetime import datetime
from typing import Union, List

from piperider_cli.assertion_engine import AssertionContext, ValidationResult
from piperider_cli.assertion_engine.types.base import BaseAssertionType


class AssertMetric(BaseAssertionType):

    def __init__(self):
        self.mapping = MetricName()

    def name(self):
        return ''

    def execute(self, context: AssertionContext):
        table = context.table
        column = context.column
        metrics = context.profiler_result

        target_metrics = metrics.get('tables', {}).get(table)
        if column:
            target_metrics = target_metrics.get('columns', {}).get(column)

        if not target_metrics:
            return context.result.fail_with_metric_not_found_error(context.table, context.column)

        context.result.name = self.mapping.get(context.metric, target_metrics.get('type'))
        context.result.expected = self.to_interval_notation(context.asserts)

        if context.metric in target_metrics:
            if target_metrics.get(context.metric) is None:
                return context.result.fail()
        else:
            return context.result.fail_with_profile_metric_not_found_error(context.table, context.column, context.metric)

        value = target_metrics.get(context.metric)
        context.result.actual = value
        if not self.assert_metric_boundary(value, context.asserts):
            return context.result.fail()

        return context.result.success()

    def validate(self, context: AssertionContext) -> ValidationResult:
        results = ValidationResult(context)

        if not self.mapping.is_exist(context.metric):
            results.errors.append(f"cannot find a metric assertion by metric '{context.metric}'")

        names = ['gte', 'lte', 'gt', 'lt', 'eq', 'ne']
        results = results.allow_only(*names) \
            .require_metric_consistency(*names)

        if context.asserts is None:
            results.errors.append(f'At least one of {names} is needed.')

        if results.errors:
            return results

        self._assert_metric_validation(context.asserts, results)

        return results

    @staticmethod
    def to_interval_notation(asserts):
        if len(asserts.keys()) == 2:
            operators = {
                'lte': ']',
                'lt': ')',
                'gte': '[',
                'gt': '('
            }
            boundary = ''
            for k, v in asserts.items():
                if k.startswith('lt'):
                    boundary += f'{v}{operators[k]}'
                else:
                    boundary = f'{operators[k]}{v}, {boundary}'
            return boundary
        else:
            operators = {
                'gt': '>',
                'gte': '≥',
                'eq': '=',
                'ne': '≠',
                'lt': '<',
                'lte': '≤'
            }
            k, v = list(asserts.items())[0]
            return f'{operators[k]} {v}'

    @staticmethod
    def assert_metric_boundary(metric: Union[int, float, str], metric_boundary: dict) -> bool:
        if isinstance(metric, str):
            metric = datetime.fromisoformat(metric)

        for op, v in metric_boundary.items():
            if isinstance(v, str):
                v = datetime.fromisoformat(v)
            if op == 'gt' and not metric > v:
                return False
            elif op == 'gte' and not metric >= v:
                return False
            elif op == 'eq' and not metric == v:
                return False
            elif op == 'ne' and not metric != v:
                return False
            elif op == 'lt' and not metric < v:
                return False
            elif op == 'lte' and not metric <= v:
                return False
        return True

    @staticmethod
    def _assert_metric_validation(metric_boundary: dict, results: ValidationResult):
        if len(metric_boundary.keys()) == 1:
            pass
        elif len(metric_boundary.keys()) == 2:
            lower = None
            upper = None
            for op, v in metric_boundary.items():
                if op == 'eq' or op == 'ne':
                    results.errors.append("Only one operator allowed if the expression contains 'eq' and 'ne'")
                    return

                if op.startswith('lt'):
                    upper = v
                elif op.startswith('gt'):
                    lower = v

            if upper is None or lower is None:
                results.errors.append('Please specified your metric upper and lower boundary')
                return

            if isinstance(upper, str) and isinstance(lower, str):
                upper = datetime.fromisoformat(upper)
                lower = datetime.fromisoformat(lower)
            if upper < lower:
                results.errors.append("The 'lt' or 'lte' value should be greater than or equal to "
                                      "the 'gt' or 'gte' value.")
        else:
            results.errors.append('The number of operator should be 1 or 2.')


class MetricName:
    def __init__(self):
        self.mapping = {}
        self.all_type = 'ALL'

        # table metric
        self._add('row_count', 'row count')
        self._add('bytes', 'volume size')
        self._add('freshness', 'freshness')
        self._add('duplicate_rows', 'duplicate row count')
        self._add('duplicate_rows_p', 'duplicate row percentage')

        self._add('total', 'row count')
        self._add('samples', 'sample count')
        self._add('samples_p', 'sample percentage')
        self._add('nulls', 'missing count')
        self._add('nulls_p', 'missing percentage')
        self._add('non_nulls', 'non null count')
        self._add('non_nulls_p', 'non null percentage')
        self._add('invalids', 'invalid count')
        self._add('invalids_p', 'invalid percentage')
        self._add('valids', 'valid count')
        self._add('valids_p', 'valid percentage')
        self._add('zeros', 'zero count', ['integer', 'numeric'])
        self._add('zeros_p', 'zero percentage', ['integer', 'numeric'])
        self._add('negatives', 'negative value count', ['integer', 'numeric'])
        self._add('negatives_p', 'negative value percentage', ['integer', 'numeric'])
        self._add('positives', 'positive value count', ['integer', 'numeric'])
        self._add('positives_p', 'positive value percentage', ['integer', 'numeric'])
        self._add('zero_length', 'zero length string count', ['string'])
        self._add('zero_length_p', 'zero length string percentage', ['string'])
        self._add('non_zero_length', 'non zero length string count', ['string'])
        self._add('non_zero_length_p', 'non zero length string percentage', ['string'])
        self._add('trues', 'true count', ['boolean'])
        self._add('trues_p', 'true percentage', ['boolean'])
        self._add('falses', 'false count', ['boolean'])
        self._add('falses_p', 'false percentage', ['boolean'])
        self._add('min', 'min', ['integer', 'numeric', 'datetime'])
        self._add('max', 'max', ['integer', 'numeric', 'datetime'])
        self._add('avg', 'average', ['integer', 'numeric'])
        self._add('sum', 'sum', ['integer', 'numeric'])
        self._add('stddev', 'standard deviation', ['integer', 'numeric'])
        self._add('min', 'min length', ['string'])
        self._add('max', 'max length', ['string'])
        self._add('avg', 'average length', ['string'])
        self._add('stddev', 'std. deviation of length', ['string'])
        self._add('distinct', 'distinct count', ['integer', 'string', 'datetime'])
        self._add('distinct_p', 'distinct percentage', ['integer', 'string', 'datetime'])
        self._add('duplicates', 'duplicate count', ['integer', 'numeric', 'string', 'datetime'])
        self._add('duplicates_p', 'duplicate percentage', ['integer', 'numeric', 'string', 'datetime'])
        self._add('non_duplicates', 'non duplicate count', ['integer', 'numeric', 'string', 'datetime'])
        self._add('non_duplicates_p', 'non duplicate percentage', ['integer', 'numeric', 'string', 'datetime'])
        self._add('min', 'min', ['integer', 'numeric'])
        self._add('p5', '5th percentile', ['integer', 'numeric'])
        self._add('p25', '25th percentile ', ['integer', 'numeric'])
        self._add('p50', 'median', ['integer', 'numeric'])
        self._add('p75', '75th percentile', ['integer', 'numeric'])
        self._add('p95', '95th percentile', ['integer', 'numeric'])
        self._add('max', 'max', ['integer', 'numeric'])

    def _add(self, field, name, col_types: List[str] = None):
        if col_types is None or len(col_types) == 0:
            if self.all_type not in self.mapping:
                self.mapping[self.all_type] = {}
            self.mapping[self.all_type][field] = name
        else:
            for t in col_types:
                if t not in self.mapping:
                    self.mapping[t] = {}
                self.mapping[t][field] = name

    def get(self, field, col_type=None):
        name = self.mapping[self.all_type].get(field)

        if name is None and col_type is not None:
            name = self.mapping[col_type].get(field)

        if name is None:
            return field

        return name

    def is_exist(self, field):
        found = False
        for type_field_mapping in self.mapping.values():
            if field in type_field_mapping.keys():
                found = True
                break

        return found



from datetime import datetime

from piperider_cli.assertion_engine import AssertionContext, AssertionResult
from piperider_cli.assertion_engine.assertion import ValidationResult
from piperider_cli.assertion_engine.types.base import BaseAssertionType


class AssertColumnMinInRange(BaseAssertionType):
    def name(self):
        return "assert_column_min_in_range"

    def execute(self, context: AssertionContext):
        return assert_column_min_in_range(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context).require('min')
        if result.has_errors():
            return result

        return result.require_range_pair('min').require_same_types('min')


class AssertColumnMaxInRange(BaseAssertionType):
    def name(self):
        return "assert_column_max_in_range"

    def execute(self, context: AssertionContext):
        return assert_column_max_in_range(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context).require('max')
        if result.has_errors():
            return result

        return result.require_range_pair('max').require_same_types('max')


class AssertColumnInRange(BaseAssertionType):
    def name(self):
        return "assert_column_in_range"

    def execute(self, context: AssertionContext):
        return assert_column_in_range(context)

    def validate(self, context: AssertionContext) -> ValidationResult:
        result = ValidationResult(context).require('range')
        if result.has_errors():
            return result

        return result.require_range_pair('range').require_same_types('range')


def assert_column_min_in_range(context: AssertionContext) -> AssertionResult:
    return _assert_column_in_range(context, target_metric='min')


def assert_column_max_in_range(context: AssertionContext) -> AssertionResult:
    return _assert_column_in_range(context, target_metric='max')


def assert_column_in_range(context: AssertionContext) -> AssertionResult:
    return _assert_column_in_range(context, target_metric='range')


def _assert_column_in_range(context: AssertionContext, **kwargs) -> AssertionResult:
    table = context.table
    column = context.column
    metrics = context.profiler_result

    table_metrics = metrics.get('tables', {}).get(table)
    if table_metrics is None:
        return context.result.fail_with_metric_not_found_error(context.table, None)

    column_metrics = table_metrics.get('columns', {}).get(column)
    if column_metrics is None:
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    # Check assertion input
    target_metric = kwargs.get('target_metric')
    values = context.asserts.get(target_metric)
    if values is None or len(values) != 2:
        return context.result.fail_with_assertion_error('Expect a range [min_value, max_value].')

    class Observed(object):
        def __init__(self, column_metrics: dict, target_metric: str):
            self.column_metrics = column_metrics
            self.target_metric = target_metric
            self.column_type = column_metrics.get('type')
            self.actual = []

            if self.target_metric == 'range':
                self.actual = [column_metrics.get('min'), column_metrics.get('max')]
            else:
                self.actual = [column_metrics.get(target_metric)]

        def is_metric_available(self):
            return [x for x in self.actual if x is None] == []

        def check_range(self, min_value, max_value):
            for metric in self.actual:
                metric = self.to_numeric(metric)
                if metric is None:
                    yield context.result.fail_with_assertion_error('Column not support range.')
                else:
                    yield min_value <= metric <= max_value

        def to_numeric(self, metric):
            if self.column_type == 'datetime':
                # TODO: check datetime format. Maybe we can leverage the format checking by YAML parser
                return datetime.strptime(metric, '%Y-%m-%d %H:%M:%S.%f')
            elif self.column_type in ['integer', 'numeric']:
                return metric
            else:
                return None

        def actual_value(self):
            if len(self.actual) == 1:
                return self.actual[0]
            return self.actual

    observed = Observed(column_metrics, target_metric)
    if not observed.is_metric_available():
        return context.result.fail_with_metric_not_found_error(context.table, context.column)

    context.result.actual = observed.actual_value()

    results = []
    for result in observed.check_range(values[0], values[1]):
        results.append(result)

    non_bools = [x for x in results if not isinstance(x, bool)]
    if non_bools:
        return non_bools[0]

    bools = [x for x in results if isinstance(x, bool)]
    if set(bools) == set([True]):
        return context.result.success()
    return context.result.fail()



from piperider_cli.error import PipeRiderError
from piperider_cli.feedback import Feedback
from . import DataSource
from .field import ListField, TextField


class UserSurveyMockDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'datasource survey', **kwargs)
        self.fields = [
            ListField('datasource_candidates',
                      description='Data Source Candidates',
                      default=[
                          'Azure Synapse Analytics',
                          'IBM Db2 Warehouse',
                          'SAP Data Warehouse Cloud',
                          'Other',
                      ]),
            TextField('other', description='Please tell what data source you want us to supported',
                      ignore=lambda answers: answers['datasource_candidates'] != 'Other'),
        ]

    # Always return None to prevent it generate the configue files
    def ask_credential(self):
        super().ask_credential()
        return None

    def validate(self):
        return False, []

    def to_database_url(self):
        return ''

    def verify_connector(self):
        return PipeRiderError('',
                              hint="Thanks for your feedback. We will improve Piperider in the future.\nPlease execute command 'piperider init' with supported data source.")

    def send_survey(self):
        datasource = self.credential.get('datasource_candidates')
        if datasource == 'Other':
            datasource = self.credential.get('other')
        Feedback.suggest_datasource(datasource)
        pass



import re
import os
import sys
import json
from abc import ABCMeta, abstractmethod
from glob import glob
from subprocess import DEVNULL, PIPE
from subprocess import Popen, check_output
from piperider_cli.error import DbtCommandNotFoundError, DbtInvocationError

import inquirer
from rich.console import Console
from rich.table import Table

console = Console()


class DbtAdaptee(metaclass=ABCMeta):

    @abstractmethod
    def set_root(self, root):
        raise NotImplementedError

    @abstractmethod
    def check(self):
        raise NotImplementedError

    @abstractmethod
    def list_resources(self, types, keys):
        raise NotImplementedError

    @abstractmethod
    def run(self, cmd_arr):
        raise NotImplementedError

    @abstractmethod
    def get_run_results(self):
        raise NotImplementedError


class DefaultDbtAdaptee(DbtAdaptee):

    def set_root(self, root):
        self.root = root

    def check(self):
        check_output(['dbt', '--version'], cwd=self.root, stderr=DEVNULL)

    def list_resources(self, types=['all'], keys=''):
        cmd_arr = ['dbt', 'list', '--output', 'json']
        for t in types:
            cmd_arr.append('--resource-type')
            cmd_arr.append(t)
        if keys:
            cmd_arr.append('--output-keys')
            cmd_arr.append(keys)

        return self._run_command(cmd_arr, stdout=PIPE, stderr=PIPE)

    def run(self, cmd_arr):
        self._run_command(cmd_arr)

    def _run_command(self, cmd_arr, stdout=None, stderr=None):
        proc = Popen(cmd_arr, stdout=stdout, stderr=stderr, cwd=self.root)
        out, err = proc.communicate()
        # Exit code ref: https://docs.getdbt.com/reference/exit-codes
        if proc.returncode == 2:
            message = "The dbt invocation completed with an error. Exit code: 2"
            if out:
                message = f"{message}\n{out.decode()}"
            raise DbtInvocationError(message)

        if out:
            return out.decode()

    def get_run_results(self):
        path = os.path.join(self.root, 'target/run_results.json')
        with open(path) as f:
            return json.load(f)


class DbtAdapter:

    def __init__(self, config, adaptee=None):
        self.config = config or {}
        self.profile = self.config.get('profile')
        self.profilesDir = self.config.get('profilesDir')
        self.projectDir = self.config.get('projectDir')
        self.target = self.config.get('target')
        self.rootPath = os.path.expanduser(self.projectDir) if self.projectDir else None

        self.adaptee: DbtAdaptee = adaptee if adaptee else DefaultDbtAdaptee()
        self.adaptee.set_root(self.rootPath)

        self.resources = []
        self.command = None
        self.error = None

        if self._check_dbt_command():
            self._list_dbt_resources()

    @staticmethod
    def search_dbt_project_path():
        exclude_patterns = ['site-packages', 'dbt_packages']
        DbtAdapter._warning_if_search_path_too_widely(os.getcwd())
        paths = glob(os.path.join(os.getcwd(), '**', 'dbt_project.yml'), recursive=True)
        for exclude_pattern in exclude_patterns:
            paths = list(filter(lambda x: exclude_pattern not in x, paths))
        dbt_project_path = None
        if not paths:
            # No dbt project found
            return dbt_project_path

        if len(paths) == 1:
            # Only one dbt project found, use it
            dbt_project_path = paths[0]
        else:
            # Multiple dbt projects found, ask user to select one
            paths = sorted(paths, key=len)
            default_index = 0
            console.print(f'\nMultiple dbt projects found. Please select the dbt project: (Default: {default_index})')
            table = Table(show_header=True, header_style="bold magenta")
            table.add_column('Index', style="bold", width=5, justify='right')
            table.add_column('Path', style="bold")
            for i, p in enumerate(paths):
                table.add_row(str(i), p)
            console.print(table)
            in_index = input('index : ').strip()
            try:
                in_index = int(in_index) if in_index else default_index
                index = int(in_index)
                dbt_project_path = paths[index]
            except Exception as e:
                console.print(f'[bold yellow]Failed to read index![/] Error: {e}')
                return None

        return dbt_project_path

    @staticmethod
    def _warning_if_search_path_too_widely(search_path):
        # Only warning user on macOS platform
        if sys.platform != "darwin":
            return

        home_dir = os.path.expanduser('~')

        if search_path in home_dir:
            console.print(
                f"[[bold yellow]Warning[/bold yellow]] Search path '{search_path}' is too widely. It will take some time to parse directories and may need extra permissions.")
            if inquirer.confirm(message="Do you still want to keep going?", default=True) is not True:
                raise KeyboardInterrupt()

    @staticmethod
    def check_dbt_command(dbt_config):
        adaptee = DefaultDbtAdaptee()
        adaptee.set_root(os.path.expanduser(dbt_config.get('projectDir', '')))
        try:
            adaptee.check()
        except Exception:
            return DbtCommandNotFoundError()

    def _check_dbt_command(self):
        if not self.is_ready():
            return False
        try:
            self.adaptee.check()
        except Exception:
            self.error = DbtCommandNotFoundError()
            return False
        return True

    def _list_dbt_resources(self):
        out = self.adaptee.list_resources()
        lines = out.split('\n')[:-1]

        # Skip lines not starts with '{', which are not message in JSON format
        self.resources = [json.loads(x) for x in lines if x.startswith('{')]

    def is_ready(self):
        # TODO: check dbt command
        if self.error:
            return False
        return None not in [self.profile, self.projectDir, self.target, self.rootPath]

    def get_error(self):
        return self.error

    def list_dbt_tables(self, default_schema):
        tables = set()
        sources = [r for r in self.resources if r['resource_type'] == 'source']
        models = [r for r in self.resources if r['resource_type'] == 'model']

        for source in sources:
            schema = source['source_name']
            schema = f'{schema}.' if schema and schema != default_schema else ''
            name = source['name']
            tables.add(f'{schema}{name}')

        for model in models:
            schema = model['config'].get('schema', default_schema)
            schema = f'{schema}.' if schema and schema != default_schema else ''
            name = model['name']
            tables.add(f'{schema}{name}')

        return list(tables)

    def set_dbt_command(self, command):
        self.command = command

    def run_dbt_command(self, table, default_schema):
        console.rule('Running dbt')
        console.print(f'[bold yellow]dbt working dir:[/bold yellow] {self.rootPath}')

        full_cmd_arr = ['dbt', self.command]
        if table:
            select = table
            if '.' in table:
                schema, name = table.split('.')[:2]
            else:
                schema, name = default_schema, table

            for resource in self.resources:
                dbt_resource_name = resource.get('name', '')
                if dbt_resource_name.upper() != name.upper():
                    continue
                if resource['resource_type'] == 'model':
                    select = name
                    break
                if resource['resource_type'] == 'source':
                    select = f"source:{resource['source_name']}.{dbt_resource_name}"
                    break
            full_cmd_arr.append('-s')
            full_cmd_arr.append(select)

        console.print(f"Execute command: {' '.join(full_cmd_arr)}")
        self.adaptee.run(full_cmd_arr)

        run_results = self.adaptee.get_run_results()

        output = []
        compatible_output = {}
        unique_tests = {}

        for result in run_results.get('results', []):
            unique_id = result.get('unique_id')
            if unique_id is None:
                continue
            unique_tests[unique_id] = dict(
                status=result.get('status'),
                failures=result.get('failures'),
                message=result.get('message'),
            )

        for resource in self.resources:
            unique_id = resource.get('unique_id')
            if unique_id not in unique_tests:
                continue

            macros = resource.get('depends_on', {}).get('macros', [])
            nodes = resource.get('depends_on', {}).get('nodes', [])
            if not nodes or not macros:
                continue

            test_method = macros[0].replace('macro.dbt.test_', '')
            node = nodes[0]
            node_type = node.split('.')[0]
            package_name = resource.get('package_name')

            is_source = 'source_' if node_type == 'source' else ''
            table_with_schema = re.sub(f'^{node_type}\.{package_name}\.', '', node)
            schema = ''
            table_name = table_with_schema
            if '.' in table_with_schema:
                schema, table_name = table_with_schema.split('.')[:2]
            is_schema = f'{schema}_' if schema != '' else ''

            pattern = f'^{is_source}{test_method}_{is_schema}{table_name}_'
            column = re.sub(pattern, '', resource['name'])

            parent_table = f'{schema}.{table_name}' if schema and schema != default_schema else table_name

            if parent_table not in compatible_output:
                compatible_output[parent_table] = dict(columns={}, tests=[])

            if column not in compatible_output[parent_table]['columns']:
                compatible_output[parent_table]['columns'][column] = []
            compatible_output[parent_table]['columns'][column].append(dict(
                id=unique_id,
                name=unique_id,
                table=parent_table,
                column=column if column != resource['name'] else None,
                status='passed' if unique_tests[unique_id]['status'] == 'pass' else 'failed',
                tags=[],
                message=unique_tests[unique_id]['message'],
                display_name=resource['name'],
                source='dbt'
            ))

            output.append(dict(
                id=unique_id,
                name=unique_id,
                table=parent_table,
                column=column if column != resource['name'] else None,
                status='passed' if unique_tests[unique_id]['status'] == 'pass' else 'failed',
                tags=[],
                message=unique_tests[unique_id]['message'],
                display_name=resource['name'],
                source='dbt'
            ))

        return output, compatible_output

    def append_descriptions(self, profile_result, default_schema):
        out = self.adaptee.list_resources(['model', 'source'], 'resource_type,description,name,columns,source_name')
        lines = out.split('\n')[:-1]
        # Skip lines not starts with '{', which are not message in JSON format
        resources = [json.loads(x) for x in lines if x.startswith('{')]
        for resource in resources:
            schema = resource.get('source_name')
            table_name = resource.get('name')
            table_desc = resource.get('description', '')

            schema = f'{schema}.' if schema and schema != default_schema else ''
            table_name = f'{schema}{table_name}'

            if table_name not in profile_result['tables']:
                continue
            if table_desc:
                profile_result['tables'][table_name]['description'] = f"{table_desc} - via DBT"

            columns = resource.get('columns', {})
            for column_name, v in columns.items():
                if column_name not in profile_result['tables'][table_name]['columns']:
                    continue
                column_desc = v.get('description', '')
                if column_desc:
                    profile_result['tables'][table_name]['columns'][column_name]['description'] = f"{column_desc} - via DBT"



import concurrent.futures
import decimal
import math
import time
from datetime import datetime, date, timezone
from typing import Union, List, Tuple

from dateutil.relativedelta import relativedelta
from sqlalchemy import MetaData, Table, Column, String, Integer, Numeric, Date, DateTime, Boolean, ARRAY, select, func, \
    distinct, case, text, literal_column, inspect
from sqlalchemy.engine import Engine, Connection
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.pool import SingletonThreadPool
from sqlalchemy.sql import FromClause
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.expression import CTE, false, true, table as table_clause, column as column_clause
from sqlalchemy.types import Float

from .event import ProfilerEventHandler, DefaultProfilerEventHandler
from ..configuration import Configuration

HISTOGRAM_NUM_BUCKET = 50


def dtof(value: Union[int, float, decimal.Decimal]) -> Union[int, float]:
    """
    dtof is helpler function to transform decimal value to float. Decimal is not json serializable type.

    :param value:
    :return:
    """
    if isinstance(value, decimal.Decimal):
        return float(value)
    return value


def format_float(val: Union[int, float]) -> str:
    """
    from the float to human-readable format.

    :param val:
    :return:
    """

    if val == 0:
        return "0"

    base = math.floor(math.log10(abs(val)))
    if base < -2:
        return f"{val:.2e}"
    elif base < 0:
        return f"{val:.3f}"
    elif base < 3:
        return f"{val:.2f}"
    elif base < 6:
        return f"{val / (10 ** 3):.1f}K"
    elif base < 9:
        return f"{val / (10 ** 6):.1f}M"
    elif base < 12:
        return f"{val / (10 ** 9):.1f}B"
    elif base < 15:
        return f"{val / (10 ** 12):.1f}T"
    else:
        return f"{val / (10 ** 12):.0f}T"


def percentage(number, total):
    if number is None:
        return None
    if not total:
        return None
    return number / total


class Profiler:
    """
    Profiler profile tables and columns by a sqlalchemy engine.
    """
    engine: Engine = None
    metadata: MetaData = None
    inspector: Inspector = None
    event_handler: ProfilerEventHandler

    def __init__(self, engine: Engine, event_handler: ProfilerEventHandler = DefaultProfilerEventHandler(),
                 config: Configuration = None):
        self.engine = engine
        self.inspector = inspect(self.engine) if self.engine else None
        self.event_handler = event_handler
        self.config = config

    def profile(self, tables: List[str] = None) -> dict:
        """
        profile all tables or specific table. With different column types, it would profile different metrics.

        The column can be categorized as these types
        - integer
        - numeric
        - string
        - datetime
        - boolean
        - Other

        :param tables: optional, the tables to profile
        :return: the profile results
        """

        profiled_tables = {}
        result = {
            "tables": profiled_tables,
        }
        self.event_handler.handle_run_start(result)

        metadata = self.metadata
        if not metadata:
            self.metadata = metadata = MetaData()
            if not tables:
                self.event_handler.handle_fetch_metadata_all_start()
                metadata.reflect(bind=self.engine)
                tables = list(self.metadata.tables.keys())
                tables = self._apply_incl_excl_tables(tables)
            else:
                tables = self._apply_incl_excl_tables(tables)
                for table in tables:
                    self.event_handler.handle_fetch_metadata_table_start(table)
                    if len(table.split('.')) == 2:
                        schema, table = table.split('.')
                        metadata.schema = schema
                    Table(table, metadata, autoload_with=self.engine)

        table_count = len(tables)
        table_index = 0
        self.event_handler.handle_run_progress(result, table_count, table_index)

        for table_name in tables:
            t = self.metadata.tables[table_name]
            tresult = self._profile_table(t)
            profiled_tables[table_name] = tresult
            table_index = table_index + 1
            self.event_handler.handle_run_progress(result, table_count, table_index)

        self.event_handler.handle_run_end(result)

        return result

    def _apply_incl_excl_tables(self, tables: List[str]) -> List[str]:
        if not self.config:
            return tables
        if self.config.includes is None and self.config.excludes is None:
            return tables

        if self.config.includes is None:
            allow_list = tables
        else:
            upper_includes = [t.upper() for t in self.config.includes]
            allow_list = [t for t in tables if t.upper() in upper_includes]

        if self.config.excludes is None:
            final_list = allow_list
        else:
            upper_excludes = [t.upper() for t in self.config.excludes]
            final_list = [t for t in allow_list if t.upper() not in upper_excludes]

        return final_list

    def _drop_unsupported_columns(self, table: Table):
        array_columns = []
        candidate_columns = []

        for column in table.columns:
            if isinstance(column.type, ARRAY):
                array_columns.append(column.name + '.')
            if any(column.name.startswith(bypass_column) for bypass_column in array_columns):
                # Skip columns under an array
                continue
            candidate_columns.append(column)
        return candidate_columns

    def _profile_table_metadata(self, result: dict, table: Table):
        row_count = created = last_altered = size_bytes = None
        with self.engine.connect() as conn:
            try:
                if self.engine.url.get_backend_name() == 'snowflake':
                    default_schema = self.inspector.default_schema_name
                    metadata_table = table_clause('TABLES', column_clause("row_count"), column_clause("created"),
                                                  column_clause("last_altered"), column_clause("bytes"),
                                                  column_clause('table_schema'), column_clause('table_name'),
                                                  schema='INFORMATION_SCHEMA')
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select([
                        metadata_columns['row_count'],
                        func.convert_timezone('UTC', metadata_columns['created']),
                        func.convert_timezone('UTC', metadata_columns['last_altered']),
                        metadata_columns['bytes']
                    ]).select_from(metadata_table).where(metadata_columns['table_schema'] == str.upper(default_schema),
                                                         metadata_columns['table_name'] == str.upper(table.name))
                    row_count, created, last_altered, size_bytes = conn.execute(stmt).fetchone()
                    # datetime object transformation
                    created = created.isoformat()
                    last_altered = last_altered.isoformat()
                elif self.engine.url.get_backend_name() == 'bigquery':
                    dataset = self.engine.url.database
                    metadata_table = table_clause(f'{dataset}.__TABLES__', column_clause("row_count"),
                                                  column_clause("creation_time"), column_clause("last_modified_time"),
                                                  column_clause("size_bytes"), column_clause('table_id'))
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select([
                        metadata_columns['row_count'],
                        metadata_columns['creation_time'],
                        metadata_columns['last_modified_time'],
                        metadata_columns['size_bytes']
                    ]).select_from(metadata_table).where(metadata_columns['table_id'] == table.name)
                    row_count, created, last_altered, size_bytes = conn.execute(stmt).fetchone()
                    # timestamp transformation
                    created = datetime.fromtimestamp(created / 1000.0, timezone.utc).isoformat()
                    last_altered = datetime.fromtimestamp(last_altered / 1000.0, timezone.utc).isoformat()
                elif self.engine.url.get_backend_name() == 'redshift':
                    metadata_table = table_clause('SVV_TABLE_INFO', column_clause("tbl_rows"),
                                                  column_clause("size"), column_clause("table"))
                    metadata_columns = {column.name: column for column in metadata_table.columns}
                    stmt = select([
                        metadata_columns['tbl_rows'],
                        metadata_columns['size'],
                    ]).select_from(metadata_table).where(metadata_columns['table'] == table.name)
                    row_count, size_mbytes = conn.execute(stmt).fetchone()
                    row_count = int(row_count)
                    size_bytes = size_mbytes * 1024
            except Exception:
                # table's metadata is optional except row_count
                pass
            finally:
                if row_count is None:
                    stmt = select([
                        func.count(),
                    ]).select_from(table)
                    row_count, = conn.execute(stmt).fetchone()

        result['row_count'] = result['samples'] = row_count
        result['samples_p'] = 1

        if self.config:
            limit = self.config.profiler_config.get('table', {}).get('limit', 0)
            if row_count > limit > 0:
                result['samples'] = limit
                result['samples_p'] = percentage(limit, row_count)

        if created:
            result['created'] = created
        if last_altered:
            result['last_altered'] = last_altered
            freshness = datetime.now(timezone.utc) - datetime.fromisoformat(last_altered)
            result['freshness'] = int(freshness.total_seconds())
        if size_bytes:
            result['bytes'] = size_bytes

    def _profile_table_duplicate_rows(self, result: dict, table: Table):
        if not self.config:
            return
        if not self.config.profiler_config.get('table', {}).get('duplicateRows'):
            return

        limit = self.config.profiler_config.get('table', {}).get('limit', 0)
        columns = [column.label(f'_{column.name}') for column in table.columns]

        with self.engine.connect() as conn:
            if self.engine.url.get_backend_name() == 'snowflake':
                if limit <= 0:
                    cte = select([func.hash(*columns).label('h')]).select_from(table).cte()
                else:
                    cte = select([func.hash(*columns).label('h')]).select_from(table).limit(limit).cte()

                cte = select([
                    cte.c.h,
                    func.count().label('c')
                ]).select_from(cte).group_by(cte.c.h).having(func.count() > 1).cte()
                stmt = select([func.sum(cte.c.c)]).select_from(cte)
                duplicate_rows, = conn.execute(stmt).fetchone()
            else:
                if limit <= 0:
                    cte = select([
                        *columns,
                        func.count().label('c')
                    ]).select_from(table).group_by(*columns).having(func.count() > 1).cte()
                else:
                    cte = select([*columns]).select_from(table).limit(limit).cte()
                    columns = [column for column in cte.columns]
                    cte = select([
                        *columns,
                        func.count().label('c')
                    ]).select_from(cte).group_by(*columns).having(func.count() > 1).cte()

                stmt = select([func.sum(cte.c.c)]).select_from(cte)
                duplicate_rows, = conn.execute(stmt).fetchone()

            samples = result['samples']
            duplicate_rows = duplicate_rows if duplicate_rows is not None else 0

            result['duplicate_rows'] = duplicate_rows
            result['duplicate_rows_p'] = percentage(duplicate_rows, samples)

    def _profile_table(self, table: Table) -> dict:
        candidate_columns = self._drop_unsupported_columns(table)
        col_index = 0
        col_count = len(candidate_columns)
        columns = {}
        result = {
            "name": table.name,
            "row_count": 0,
            "samples": 0,
            "samples_p": None,
            "col_count": col_count,
            "duplicate_rows": None,
            "duplicate_rows_p": None,
            "columns": columns
        }

        if not self.engine:
            # unittest case
            return result

        self.event_handler.handle_table_start(result)
        profile_start = time.perf_counter()

        # Profile table metrics
        self._profile_table_metadata(result, table)
        self._profile_table_duplicate_rows(result, table)
        self.event_handler.handle_table_progress(result, col_count, col_index)

        # Profile columns
        samples_p = result['samples_p']
        if isinstance(self.engine.pool, SingletonThreadPool):
            for column in candidate_columns:
                columns[column.name] = self._profile_column(table, column)
                columns[column.name]['total'] = result['row_count']
                columns[column.name]['samples_p'] = samples_p
                col_index = col_index + 1
                self.event_handler.handle_table_progress(result, col_count, col_index)
            profile_end = time.perf_counter()
            duration = profile_end - profile_start
            result["profile_duration"] = f"{duration:.2f}"
            result["elapsed_milli"] = int(duration * 1000)

            self.event_handler.handle_table_end(result)
        else:
            for column in candidate_columns:
                columns[column.name] = None
            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                future_to_profile = {executor.submit(self._profile_column, table, column): column for column in
                                     candidate_columns}
                try:
                    for future in concurrent.futures.as_completed(future_to_profile):
                        column = future_to_profile[future]
                        try:
                            data = future.result()
                        except Exception as exc:
                            raise exc
                        else:
                            columns[column.name] = data
                            columns[column.name]['total'] = result['row_count']
                            columns[column.name]['samples_p'] = samples_p
                            col_index = col_index + 1
                            self.event_handler.handle_table_progress(result, col_count, col_index)
                finally:
                    profile_end = time.perf_counter()
                    duration = profile_end - profile_start
                    result["profile_duration"] = f"{duration:.2f}"
                    result["elapsed_milli"] = int(duration * 1000)

                    self.event_handler.handle_table_end(result)

        return result

    def _profile_column(self, table: Table, column: Column) -> dict:
        profiler_config = self.config.profiler_config if self.config else {}
        if isinstance(column.type, String):
            # VARCHAR
            # CHAR
            # TEXT
            # CLOB
            generic_type = "string"
            profiler = StringColumnProfiler(self.engine, profiler_config, table, column)
        elif isinstance(column.type, Integer):
            # INTEGER
            # BIGINT
            # SMALLINT
            generic_type = "integer"
            profiler = NumericColumnProfiler(self.engine, profiler_config, table, column, is_integer=True)
        elif isinstance(column.type, Numeric):
            # NUMERIC
            # DECIMAL
            # FLOAT
            generic_type = "numeric"
            profiler = NumericColumnProfiler(self.engine, profiler_config, table, column, is_integer=False)
        elif isinstance(column.type, Date) or isinstance(column.type, DateTime) or \
            (self.engine.url.get_backend_name() == 'snowflake' and str(column.type).startswith('TIMESTAMP')):
            # DATE
            # DATETIME
            # TIMEZONE_NTZ
            generic_type = "datetime"
            profiler = DatetimeColumnProfiler(self.engine, profiler_config, table, column)
        elif isinstance(column.type, Boolean):
            # BOOLEAN
            generic_type = "boolean"
            profiler = BooleanColumnProfiler(self.engine, profiler_config, table, column)
        else:
            generic_type = "other"
            profiler = BaseColumnProfiler(self.engine, profiler_config, table, column)

        result = {
            "name": column.name,
            "type": generic_type,
            "schema_type": str(column.type)
        }

        self.event_handler.handle_column_start(table.name, result)

        profile_start = time.perf_counter()
        profile_result = profiler.profile()
        profile_end = time.perf_counter()
        duration = profile_end - profile_start

        result.update(profile_result)
        result["profile_duration"] = f"{duration:.2f}"
        result["elapsed_milli"] = int(duration * 1000)

        self.event_handler.handle_column_end(table.name, result)

        return result


class BaseColumnProfiler:
    """
    The base class of the column profiler. It will automatically profile the metrics according to the schema type
    """

    engine: Engine = None
    config: dict = None
    table: Table = None
    column: Column = None

    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        self.engine = engine
        self.config = config
        self.table = table
        self.column = column

    def _get_database_backend(self) -> str:
        """
        Helper function to return the sqlalchemy engine backend
        :return:
        """
        return self.engine.url.get_backend_name()

    def _get_limited_table_cte(self):
        t = self.table
        c = self.column
        if not self.config:
            return t, c

        limit = self.config.get('table', {}).get('limit', 0)
        if limit <= 0:
            return t, c
        else:
            cte = select([c.label('c')]).select_from(t).limit(limit).cte()
            return cte, cte.c.c

    def _get_table_cte(self) -> CTE:
        """
        Get the CTE to normalize the
        - table name
        - column name as column "c"
        - (Optional) Remove the invalid data.

        Columns
        - "c": the transformed valid to use data
            null: if the column value is null or invalid
            otherwise: orginal column value or transformed value.
        - "orig": the original column

        so that

        valid       = count(c)
        non_nulls   = count(orig)
        invalids    = non_nulls - valid

        :return: CTE
        """
        t, c = self._get_limited_table_cte()

        return select([c.label("c")]).select_from(t).cte()

    def profile(self) -> dict:
        """
        Profile a column

        :return: the profiling result. The result dict is json serializable
        """

        with self.engine.connect() as conn:
            cte = self._get_table_cte()
            stmt = select([
                func.count().label("_total"),
                func.count(cte.c.c).label("_non_nulls"),
            ])
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, = result
            _nulls = _total - _non_nulls
            _valid = _non_nulls

            return {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valid,
                'valids_p': percentage(_valid, _total),
                'invalids': 0,
                'invalids_p': 0,
                'distribution': None,
            }


class StringColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select([
                c.label("c"),
                c.label("orig")
            ]).select_from(t).cte()
        else:
            cte = select([
                case(
                    [(func.typeof(c) == 'blob', None)],
                    else_=c
                ).label("c"),
                c.label("orig"),
            ]).select_from(t).cte()
        cte = select([
            cte.c.c,
            func.length(cte.c.c).label("len"),
            cte.c.orig
        ]).select_from(cte).cte()
        cte = select([
            cte.c.c,
            cte.c.len,
            case([(cte.c.len == 0, 1)], else_=None).label("zero_length"),
            cte.c.orig
        ]).select_from(cte).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            columns = [
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.zero_length).label("_zero_length"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.avg(cte.c.len).label("_avg"),
                func.min(cte.c.len).label("_min"),
                func.max(cte.c.len).label("_max"),
            ]

            if self._get_database_backend() == 'sqlite':
                columns.append((func.count(cte.c.len) * func.sum(
                    func.cast(cte.c.len, Float) * func.cast(cte.c.len, Float)) - func.sum(cte.c.len) * func.sum(
                    cte.c.len)) / ((func.count(cte.c.len) - 1) * func.count(cte.c.len)).label('_variance'))
                stmt = select(columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zero_length, _distinct, _avg, _min, _max, _variance = result
                _stddev = None
                if _variance is not None:
                    _stddev = math.sqrt(_variance)
            else:
                columns.append(func.stddev(cte.c.len).label("_stddev"))
                stmt = select(columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zero_length, _distinct, _avg, _min, _max, _stddev = result

            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _non_zero_length = _valids - _zero_length
            _min = dtof(_min)
            _max = dtof(_max)
            _avg = dtof(_avg)
            _stddev = dtof(_stddev)

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'zero_length': _zero_length,
                'zero_length_p': percentage(_zero_length, _total),
                'non_zero_length': _non_zero_length,
                'non_zero_length_p': percentage(_non_zero_length, _total),

                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min,
                'min_length': _min,
                'max': _max,
                'max_length': _max,
                'avg': _avg,
                'avg_length': _avg,
                'stddev': _stddev,
                'stddev_length': _stddev,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # top k
            topk = None
            if _valids > 0:
                topk = profile_topk(conn, cte.c.c)
            result['topk'] = topk

            # histogram of string length
            histogram = None
            if _valids > 0:
                histogram = profile_histogram(conn, cte, cte.c.len, _min, _max, True)
            result['histogram'] = histogram
            result['histogram_length'] = histogram

            # deprecated
            result['distribution'] = {
                "type": "topk",
                "labels": topk["values"],
                "counts": topk["counts"],
            } if topk else None

            return result


class NumericColumnProfiler(BaseColumnProfiler):
    is_integer: bool

    def __init__(self, engine: Engine, config: dict, table: Table, column: Column, is_integer: bool):
        super().__init__(engine, config, table, column)
        self.is_integer = is_integer

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select([
                c.label("c"),
                c.label("orig")
            ]).select_from(t).cte()
        else:
            cte = select([
                case(
                    [(func.typeof(c) == 'text', None),
                     (func.typeof(c) == 'blob', None)],
                    else_=c
                ).label("c"),
                c.label("orig")
            ]).select_from(t).cte(name="T")
        cte = select([
            cte.c.c,
            case([(cte.c.c == 0, 1)], else_=None).label("zero"),
            case([(cte.c.c < 0, 1)], else_=None).label("negative"),
            cte.c.orig
        ]).select_from(cte).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            columns = [
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.zero).label("_zeros"),
                func.count(cte.c.negative).label("_negatives"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.sum(func.cast(cte.c.c, Float)).label("_sum"),
                func.avg(cte.c.c).label("_avg"),
                func.min(cte.c.c).label("_min"),
                func.max(cte.c.c).label("_max"),
            ]

            if self._get_database_backend() == 'sqlite':
                columns.append((func.count(cte.c.c) * func.sum(
                    func.cast(cte.c.c, Float) * func.cast(cte.c.c, Float)) - func.sum(cte.c.c) * func.sum(cte.c.c)) / (
                                   (func.count(cte.c.c) - 1) * func.count(cte.c.c)).label('_variance'))
                stmt = select(columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zeros, _negatives, _distinct, _sum, _avg, _min, _max, _variance = result
                _stddev = None
                if _variance is not None:
                    _stddev = math.sqrt(_variance)
            else:
                columns.append(func.stddev(cte.c.c).label("_stddev"))
                stmt = select(columns)
                result = conn.execute(stmt).fetchone()
                _total, _non_nulls, _valids, _zeros, _negatives, _distinct, _sum, _avg, _min, _max, _stddev = result

            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _positives = _valids - _zeros - _negatives
            _sum = dtof(_sum)
            _min = dtof(_min)
            _max = dtof(_max)
            _avg = dtof(_avg)
            _stddev = dtof(_stddev)

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'zeros': _zeros,
                'zeros_p': percentage(_zeros, _total),
                'negatives': _negatives,
                'negatives_p': percentage(_negatives, _total),
                'positives': _positives,
                'positives_p': percentage(_positives, _total),

                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min,
                'max': _max,
                'sum': _sum,
                'avg': _avg,
                'stddev': _stddev,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # histogram
            histogram = None
            if _valids > 0:
                histogram = profile_histogram(conn, cte, cte.c.c, _min, _max, self.is_integer)
            result['histogram'] = histogram

            # quantile
            quantile = {}
            if _valids > 0:
                quantile = self._profile_quantile(conn, cte, cte.c.c, _valids)
            result.update({
                'p5': quantile.get('p5'),
                'p25': quantile.get('p25'),
                'p50': quantile.get('p50'),
                'p75': quantile.get('p75'),
                'p95': quantile.get('p95'),
            })

            # top k (integer only)
            if self.is_integer:
                topk = None
                if _valids > 0:
                    topk = profile_topk(conn, cte.c.c)
                result["topk"] = topk

            # deprecated
            result["distribution"] = {
                "type": "histogram",
                "labels": histogram["labels"],
                "counts": histogram["counts"],
                "bin_edges": histogram["bin_edges"],
            } if histogram else None

            return result

    def _profile_quantile_via_window_function(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ):
        # with t as (
        #   select
        #     column as c,
        #     ntile(20) over (order by column) as n
        #   from table
        # )
        # select n, min(c) from t group by n order by n
        n_bucket = total if total < 100 else 100
        t = select([
            column.label("c"),
            func.ntile(n_bucket).over(order_by=column).label("n")
        ]).where(column.isnot(None)).select_from(table).cte()
        stmt = select([t.c.n, func.min(t.c.c)]).group_by(t.c.n).order_by(t.c.n)
        result = conn.execute(stmt)
        quantile = []
        for row in result:
            n, v = row
            quantile.append(v)
        return {
            'p5': dtof(quantile[5 * n_bucket // 100]),
            'p25': dtof(quantile[25 * n_bucket // 100]),
            'p50': dtof(quantile[50 * n_bucket // 100]),
            'p75': dtof(quantile[75 * n_bucket // 100]),
            'p95': dtof(quantile[95 * n_bucket // 100]),
        }

    def _profile_quantile_via_query_one_by_one(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ):
        # Query for each quantile
        def ntile(n):
            offset = n * total // 100

            stmt = select([
                column
            ]).select_from(
                table
            ).where(
                column.isnot(None)
            ).order_by(
                column
            ).offset(
                offset
            ).limit(1)
            result, = conn.execute(stmt).fetchone()
            return dtof(result)

        return {
            'p5': ntile(5),
            'p25': ntile(25),
            'p50': ntile(50),
            'p75': ntile(75),
            'p95': ntile(95),
        }

    def _profile_quantile(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        total: int
    ) -> dict:
        """

        :param conn:
        :param table: a
        :param column:
        :param total:
        :return:
        """

        if self._get_database_backend() == 'sqlite':
            import sqlite3
            version = sqlite3.sqlite_version.split(".")

            major = int(version[0]) if len(version) >= 2 else 0
            minor = int(version[1]) if len(version) >= 2 else 0

            if major > 3 or (major == 3 and minor >= 25):
                # use window function if sqlite version >= 3.25.0
                # see https://www.sqlite.org/windowfunctions.html

                return self._profile_quantile_via_window_function(conn, table, column, total)
            else:
                return self._profile_quantile_via_query_one_by_one(conn, table, column, total)
        elif self._get_database_backend() == 'duckdb':
            selects = [
                func.approx_quantile(column, literal_column(f"{percentile}")) for percentile in
                [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        elif self._get_database_backend() == 'bigquery':
            # BigQuery does not support WITHIN, change to use over
            #   Ref: https://github.com/great-expectations/great_expectations/blob/develop/great_expectations/dataset/sqlalchemy_dataset.py#L1019:9
            selects = [
                func.percentile_disc(column, percentile).over() for percentile in [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        elif self._get_database_backend() == 'redshift':
            # ref: https://docs.aws.amazon.com/redshift/latest/dg/r_APPROXIMATE_PERCENTILE_DISC.html
            selects = [
                func.approximate_percentile_disc(percentile).within_group(column) for percentile in
                [0.05, 0.25, 0.5, 0.75, 0.95]
            ]
        else:
            # https://docs.sqlalchemy.org/en/14/core/functions.html#sqlalchemy.sql.functions.percentile_disc
            #
            # select
            #     percentile_disc(0.05) within group (order by column),
            #     percentile_disc(0.25) within group (order by column),
            #     percentile_disc(0.5) within group (order by column),
            #     percentile_disc(0.75) within group (order by column),
            #     percentile_disc(0.95) within group (order by column)
            # from table
            selects = [
                func.percentile_disc(percentile).within_group(column) for percentile in [0.05, 0.25, 0.5, 0.75, 0.95]
            ]

        stmt = select(selects).select_from(table)
        result = conn.execute(stmt).fetchone()
        return {
            'p5': dtof(result[0]),
            'p25': dtof(result[1]),
            'p50': dtof(result[2]),
            'p75': dtof(result[3]),
            'p95': dtof(result[4]),
        }

    def _profile_histogram(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        min: Union[int, float],
        max: Union[int, float],
        is_integer: bool,
        num_buckets: int = HISTOGRAM_NUM_BUCKET
    ) -> dict:
        if is_integer:
            # min=0, max=50, num_buckets=50  => interval=1, num_buckets=51
            # min=0, max=70, num_buckets=50  => interval=2, num_buckets=36
            # min=0, max=100, num_buckets=50 => interval=2, num_buckets=51
            interval = math.ceil((max - min) / num_buckets) if max > min else 1
            num_buckets = math.ceil((max - min + 1) / interval)
        else:
            interval = (max - min) / num_buckets if max > min else 1

        cases = []
        for i in range(num_buckets):
            bound = min + interval * (i + 1)
            if i != num_buckets - 1:
                cases += [(column < bound, i)]
            else:
                cases += [(column < bound + interval / 100, i)]

        cte_with_bucket = select([
            column.label("c"),
            case(cases, else_=None).label("bucket")
        ]).select_from(
            table
        ).where(
            column.isnot(None)
        ).cte()

        stmt = select([
            cte_with_bucket.c.bucket,
            func.count().label("_count")
        ]).group_by(
            cte_with_bucket.c.bucket
        ).order_by(
            cte_with_bucket.c.bucket
        )

        result = conn.execute(stmt)

        counts = []
        labels = []
        bin_edges = []
        for i in range(num_buckets):
            if is_integer:
                start = min + i * interval
                end = min + (i + 1) * interval
                if interval == 1:
                    label = f"{start}"
                else:
                    label = f"{start} _ {end}"
            else:
                if interval >= 1:
                    start = min + i * interval
                    end = min + (i + 1) * interval
                else:
                    start = min + i / (1 / interval)
                    end = min + (i + 1) / (1 / interval)

                label = f"{format_float(start)} _ {format_float(end)}"

            labels.append(label)
            counts.append(0)
            bin_edges.append(start)
            if i == num_buckets - 1:
                bin_edges.append(end)

        for row in result:
            _bucket, v = row
            if _bucket is None:
                continue
            counts[int(_bucket)] = v
        return {
            "type": "histogram",
            "labels": labels,
            "counts": counts,
            "bin_edges": bin_edges,
        }


class DatetimeColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select([
                c.label("c"),
                c.label("orig")
            ]).select_from(t).cte()
        else:
            cte = select([
                case(
                    [(func.typeof(c) == 'text', func.datetime(c)),
                     (func.typeof(c) == 'integer', func.datetime(c, 'unixepoch')),
                     (func.typeof(c) == 'real', func.datetime(c, 'unixepoch'))],
                    else_=None
                ).label("c"),
                c.label("orig"),
            ]).select_from(t).cte()
        return cte

    def profile(self):
        with self.engine.connect() as conn:
            cte = self._get_table_cte()

            stmt = select([
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(distinct(cte.c.c)).label("_distinct"),
                func.min(cte.c.c).label("_min"),
                func.max(cte.c.c).label("_max"),
            ])
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, _valids, _distinct, _min, _max = result
            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids

            if self._get_database_backend() == 'sqlite':
                if isinstance(self.column.type, Date):
                    _min = datetime.fromisoformat(_min).date() if isinstance(_min, str) else _min
                    _max = datetime.fromisoformat(_max).date() if isinstance(_max, str) else _max
                else:
                    _min = datetime.fromisoformat(_min) if isinstance(_min, str) else _min
                    _max = datetime.fromisoformat(_max) if isinstance(_max, str) else _max

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),
                'min': _min.isoformat() if _min is not None else None,
                'max': _max.isoformat() if _max is not None else None,
            }

            # uniqueness
            _non_duplicates = profile_non_duplicate(conn, cte, cte.c.c)
            _duplicates = _valids - _non_duplicates
            result.update({
                "duplicates": _duplicates,
                "duplicates_p": percentage(_duplicates, _valids),
                "non_duplicates": _non_duplicates,
                "non_duplicates_p": percentage(_non_duplicates, _valids),
            })

            # histogram
            histogram = None
            _type = None
            if _min and _max:
                histogram, _type = self._profile_histogram(conn, cte, cte.c.c, _min, _max)
            result['histogram'] = histogram

            # deprecated
            result["distribution"] = {
                "type": _type,
                "labels": histogram["labels"],
                "counts": histogram["counts"],
                "bin_edges": histogram["bin_edges"],
            } if histogram else None

            return result

    def _profile_histogram(
        self,
        conn: Connection,
        table: FromClause,
        column: ColumnClause,
        min: Union[date, datetime],
        max: Union[date, datetime]
    ) -> Tuple[dict, str]:
        """
        Profile the histogram of a datetime column. There are three way to create bins of the histogram


        :param conn:
        :param table:
        :param column:
        :param min:
        :param max:
        :return:
        """

        # if self._get_database_backend() == 'sqlite':
        #     min = datetime.fromisoformat(min).date() if min is not None else min
        #     max = datetime.fromisoformat(max).date() if max is not None else max
        # else:
        #     if isinstance(min, datetime):
        #         min = min.date()
        #     if isinstance(max, datetime):
        #         max = max.date()

        _type = None
        histogram = {
            "labels": [],
            "counts": [],
            "bin_edges": [],
        }

        def date_trunc(*args):
            if self._get_database_backend() == 'sqlite':
                if args[0] == "YEAR":
                    return func.strftime("%Y-01-01", args[1])
                elif args[0] == "MONTH":
                    return func.strftime("%Y-%m-01", args[1])
                else:
                    return func.strftime("%Y-%m-%d", args[1])
            elif self._get_database_backend() == 'bigquery':
                date_expression = args[1]
                date_part = args[0]
                return func.date_trunc(date_expression, text(date_part))
            else:
                return func.date_trunc(*args)

        days_delta = (max - min).days
        if days_delta > 365 * 4:
            _type = "yearly"
            dmin = date(min.year, 1, 1)
            dmax = date(max.year, 1, 1) + relativedelta(years=+1)
            interval_years = math.ceil((dmax.year - dmin.year) / 50)
            interval = relativedelta(years=+interval_years)
            num_buckets = math.ceil((dmax.year - dmin.year) / interval.years)
            cte = select([date_trunc("YEAR", column).label("d")]).select_from(table).cte()
        elif days_delta > 60:
            _type = "monthly"
            interval = relativedelta(months=+1)
            dmin = date(min.year, min.month, 1)
            dmax = date(max.year, max.month, 1) + interval
            period = relativedelta(dmax, dmin)
            num_buckets = (period.years * 12 + period.months)
            cte = select([date_trunc("MONTH", column).label("d")]).select_from(table).cte()
        else:
            _type = "daily"
            interval = relativedelta(days=+1)
            dmin = date(min.year, min.month, min.day)
            dmax = date(max.year, max.month, max.day) + interval
            num_buckets = (dmax - dmin).days
            cte = select([date_trunc("DAY", column).label("d")]).select_from(table).cte()

        stmt = select([
            cte.c.d,
            func.count(cte.c.d).label("_count")
        ]).group_by(
            cte.c.d
        ).order_by(
            cte.c.d
        )

        result = conn.execute(stmt)

        for i in range(num_buckets):
            label = f"{dmin + i * interval} - {dmin + (i + 1) * interval}"
            histogram["labels"].append(label)
            histogram["bin_edges"].append(str(dmin + i * interval))
            histogram["counts"].append(0)
        histogram["bin_edges"].append(str(dmin + num_buckets * interval))

        for row in result:
            date_truncated, v = row
            if date_truncated is None:
                continue
            elif isinstance(date_truncated, str):
                date_truncated = date.fromisoformat(date_truncated)
            elif isinstance(date_truncated, datetime):
                date_truncated = date_truncated.date()

            for i in range(num_buckets):
                date_edge = date.fromisoformat(histogram["bin_edges"][i + 1])
                if date_truncated < date_edge:
                    histogram["counts"][i] += v
                    break

        return histogram, _type


class BooleanColumnProfiler(BaseColumnProfiler):
    def __init__(self, engine: Engine, config: dict, table: Table, column: Column):
        super().__init__(engine, config, table, column)

    def _get_table_cte(self) -> CTE:
        t, c = self._get_limited_table_cte()
        if self._get_database_backend() != 'sqlite':
            cte = select([
                c.label("c"),
                c.label("orig")
            ]).select_from(t).cte()
        else:
            cte = select([
                case(
                    [(c == true(), c),
                     (c == false(), c)],
                    else_=None
                ).label("c"),
                c.label("orig"),
            ]).cte()
        cte = select([
            cte.c.c,
            case([(cte.c.c == true(), 1)], else_=None).label("true_count"),
            cte.c.orig
        ]).select_from(cte).cte()
        return cte

    def profile(self):
        cte = self._get_table_cte()

        with self.engine.connect() as conn:
            stmt = select([
                func.count().label("_total"),
                func.count(cte.c.orig).label("_non_nulls"),
                func.count(cte.c.c).label("_valids"),
                func.count(cte.c.true_count).label("_trues"),
                func.count(distinct(cte.c.c)).label("_distinct"),
            ]).select_from(cte)
            result = conn.execute(stmt).fetchone()
            _total, _non_nulls, _valids, _trues, _distinct = result
            _nulls = _total - _non_nulls
            _invalids = _non_nulls - _valids
            _falses = _valids - _trues

            result = {
                'total': None,
                'samples': _total,
                'samples_p': None,
                'non_nulls': _non_nulls,
                'non_nulls_p': percentage(_non_nulls, _total),
                'nulls': _nulls,
                'nulls_p': percentage(_nulls, _total),
                'valids': _valids,
                'valids_p': percentage(_valids, _total),
                'invalids': _invalids,
                'invalids_p': percentage(_invalids, _total),
                'trues': _trues,
                'trues_p': percentage(_trues, _total),
                'falses': _falses,
                'falses_p': percentage(_falses, _total),
                'distinct': _distinct,
                'distinct_p': percentage(_distinct, _valids),

                # deprecated
                'distribution': {
                    'type': "topk",
                    'labels': ["False", "True"],
                    'counts': [_falses, _trues]
                }
            }

            return result


def profile_topk(conn, expr, k=50) -> dict:
    stmt = select([
        expr,
        func.count().label("_count")
    ]).where(
        expr.isnot(None)
    ).group_by(
        expr
    ).order_by(
        func.count().desc(),
    ).limit(k)
    result = conn.execute(stmt)

    topk = {
        "values": [],
        "counts": [],
    }
    for row in result:
        k, v = row
        if k is not None:
            k = str(k)
        topk["values"].append(k)
        topk["counts"].append(v)
    return topk


def profile_histogram(
    conn: Connection,
    table: FromClause,
    column: ColumnClause,
    min: Union[int, float],
    max: Union[int, float],
    is_integer: bool,
    num_buckets: int = HISTOGRAM_NUM_BUCKET
) -> dict:
    if is_integer:
        # min=0, max=50, num_buckets=50  => interval=1, num_buckets=51
        # min=0, max=70, num_buckets=50  => interval=2, num_buckets=36
        # min=0, max=100, num_buckets=50 => interval=2, num_buckets=51
        interval = math.ceil((max - min) / num_buckets) if max > min else 1
        num_buckets = math.ceil((max - min + 1) / interval)
    else:
        interval = (max - min) / num_buckets if max > min else 1

    cases = []
    for i in range(num_buckets):
        bound = min + interval * (i + 1)
        if i != num_buckets - 1:
            cases += [(column < bound, i)]
        else:
            cases += [(column < bound + interval / 100, i)]

    cte_with_bucket = select([
        column.label("c"),
        case(cases, else_=None).label("bucket")
    ]).select_from(
        table
    ).where(
        column.isnot(None)
    ).cte()

    stmt = select([
        cte_with_bucket.c.bucket,
        func.count().label("_count")
    ]).group_by(
        cte_with_bucket.c.bucket
    ).order_by(
        cte_with_bucket.c.bucket
    )

    result = conn.execute(stmt)

    counts = []
    labels = []
    bin_edges = []
    for i in range(num_buckets):
        if is_integer:
            start = min + i * interval
            end = min + (i + 1) * interval
            if interval == 1:
                label = f"{start}"
            else:
                label = f"{start} _ {end}"
        else:
            if interval >= 1:
                start = min + i * interval
                end = min + (i + 1) * interval
            else:
                start = min + i / (1 / interval)
                end = min + (i + 1) / (1 / interval)

            label = f"{format_float(start)} _ {format_float(end)}"

        labels.append(label)
        counts.append(0)
        bin_edges.append(start)
        if i == num_buckets - 1:
            bin_edges.append(end)

    for row in result:
        _bucket, v = row
        if _bucket is None:
            continue
        counts[int(_bucket)] = v
    return {
        "labels": labels,
        "counts": counts,
        "bin_edges": bin_edges,
    }


def profile_non_duplicate(
    conn: Connection,
    table: FromClause,
    column: ColumnClause,
) -> int:
    # with t as (
    #     select count(column) as c
    # from table
    # group by column
    # having c == 1
    # )
    # select
    # count(c) as non_duplicate
    # from t;
    cte = select([
        func.count(column).label("non_duplicates")
    ]).select_from(
        table
    ).where(
        column.isnot(None)
    ).group_by(
        column
    ).having(
        func.count(column) == 1
    ).cte()

    stmt = select([func.count(cte.c.non_duplicates)]).select_from(cte)
    non_duplicates, = conn.execute(stmt).fetchone()
    return non_duplicates



from rich.color import Color
from rich.progress import Progress, Column, TextColumn, BarColumn, TimeElapsedColumn, MofNCompleteColumn
from rich.style import Style


class AssertionEventHandler:
    def handle_assertion_start(self, assertions):
        pass

    def handle_assertion_end(self, results, exceptions):
        pass

    def handle_execution_start(self, assertion_context):
        pass

    def handle_execution_end(self, assertion_result):
        pass


class DefaultAssertionEventHandler(AssertionEventHandler):

    def __init__(self):
        subject_column = TextColumn("{task.fields[test_subject]}")
        assert_column = TextColumn("{task.fields[display_name]}", style='green')
        bar_column = BarColumn(bar_width=60, pulse_style=Style.from_color(Color.from_rgb(244, 164, 96)))
        mofn_column = MofNCompleteColumn(table_column=Column(width=5, justify="right"))
        time_elapsed_column = TimeElapsedColumn()
        self.progress = Progress(subject_column, assert_column, bar_column, mofn_column, time_elapsed_column)
        self.task_id = None

    def handle_assertion_start(self, assertions):
        self.progress.start()
        fields = dict(display_name='', test_subject='')
        self.task_id = self.progress.add_task('assertions', total=len(assertions), **fields)

    def handle_assertion_end(self, results, exceptions):
        self.progress.update(self.task_id, **dict(display_name='', test_subject='Test completed'))
        self.progress.stop()

    def handle_execution_start(self, assertion_context):
        display_name = assertion_context.name if assertion_context.name is not None else assertion_context.metric
        test_subject = assertion_context.table if assertion_context.table else '-'
        if assertion_context.column:
            test_subject = f'{test_subject}.{assertion_context.column}'
        self.progress.update(self.task_id, **dict(display_name=display_name, test_subject=test_subject))

    def handle_execution_end(self, assertion_result):
        self.progress.update(self.task_id, advance=1, **dict(display_name=''))



class ProfilerEventHandler:
    def handle_run_start(self, run_result):
        pass

    def handle_run_end(self, run_result):
        pass

    def handle_run_progress(self, run_result, total, completed):
        pass

    def handle_fetch_metadata_all_start(self):
        pass

    def handle_fetch_metadata_table_start(self, table_name):
        pass

    def handle_table_start(self, table_result):
        pass

    def handle_table_progress(self, table_result, total, completed):
        pass

    def handle_table_end(self, table_result):
        pass

    def handle_column_start(self, table_name, column_result):
        pass

    def handle_column_end(self, table_name, column_result):
        pass


class DefaultProfilerEventHandler(ProfilerEventHandler):
    table_completed = 0
    table_total = 0
    col_completed = 0
    col_total = 0

    def handle_run_start(self, run_result):
        print("Start profiling")

    def handle_run_progress(self, run_result, total, completed):
        self.table_total = total
        self.table_completed = completed

    def handle_run_end(self, run_result):
        pass

    def handle_fetch_metadata_all_start(self):
        print("fetching metadata")

    def handle_fetch_metadata_table_start(self, table_name):
        print(f"fetching metadata for table '{table_name}'")

    def handle_table_start(self, table_result):
        print(f"[{self.table_completed + 1}/{self.table_total}] profiling [{table_result['name']}] ", end='', flush=True)

    def handle_table_progress(self, table_result, total, completed):
        self.col_total = total
        self.col_completed = completed
        if completed == 0:
            print(f"\r[{self.table_completed + 1}/{self.table_total}] profiling [{table_result['name']}] rows={table_result['row_count']}")

    def handle_table_end(self, table_result):
        pass

    def handle_column_start(self, table_name, column_result):
        print(
            f"    [{self.col_completed + 1}/{self.col_total}] profiling [{table_name}.{column_result['name']}] type={column_result['schema_type']}",
            end='', flush=True)

    def handle_column_end(self, table_name, column_result):
        print(
            f"\r    [{self.col_completed + 1}/{self.col_total}] profiling [{table_name}.{column_result['name']}] type={column_result['schema_type']} [{column_result['elapsed_milli']}ms]")



from piperider_cli.error import PipeRiderConnectorError, PipeRiderCredentialFieldError
from . import DataSource
from .field import TextField, PasswordField


class SnowflakeDataSource(DataSource):
    def __init__(self, name, **kwargs):
        super().__init__(name, 'snowflake', **kwargs)
        # self.fields = ["account", "user", "password", "database", "warehouse", "schema"]
        self.fields = [
            TextField('account', description='Account'),
            TextField('user', description='Username'),
            PasswordField('password', description='Password'),
            TextField('role', description='Role', optional=True),
            TextField('database', description='Database'),
            TextField('warehouse', description='Warehouse'),
            TextField('schema', default='PUBLIC', description='Schema'),
        ]
        self._connect_timeout = 5

    def validate(self):
        if self.type_name != 'snowflake':
            raise ValueError('type name should be snowflake')
        return self._validate_required_fields()

    def to_database_url(self):
        credential = self.credential
        account = credential.get('account')
        password = credential.get('password')
        user = credential.get('user')
        database = credential.get('database')
        schema = credential.get('schema')
        warehouse = credential.get('warehouse')
        role = credential.get('role')
        authenticator = credential.get('authenticator')
        from snowflake.sqlalchemy.snowdialect import SnowflakeDialect
        from snowflake.sqlalchemy import URL

        SnowflakeDialect.supports_statement_cache = True
        db_parameters = {
            "account": account,
            "user": user,
            "password": password,
            "database": database,
            "schema": schema,
            "warehouse": warehouse
        }

        if role:
            db_parameters["role"] = role

        if authenticator:
            if authenticator not in ['snowflake', 'username_password_mfa']:
                raise PipeRiderCredentialFieldError('authenticator', 'The authentication method is not supported')
            db_parameters["authenticator"] = authenticator

        return URL(**db_parameters)

    def engine_args(self):
        return dict(connect_args={'connect_timeout': self._connect_timeout})

    def verify_connector(self):
        try:
            import snowflake.connector
            import snowflake.sqlalchemy
            # do nothing when everything is ok
            return None
        except Exception as e:
            return PipeRiderConnectorError(str(e), 'snowflake')



import configparser
import os
from os.path import basename, splitext
from urllib.parse import urlparse

import requests
from rich.console import Console
from sqlalchemy.exc import DBAPIError

from piperider_cli.error import PipeRiderConnectorError, AwsCredentialsError, AwsUnExistedS3Bucket, \
    PipeRiderDataBaseConnectionError, PipeRiderDataBaseEncodingError
from . import DataSource
from .field import PathField, TextField

AWS_CREDENTIAL_PATH = os.path.expanduser('~/.aws/credentials')


def _parquet_path_validate_func(answer, current) -> bool:
    if current is None or current == '':
        return False

    if current.startswith('http://') or current.startswith('https://'):
        try:
            # Check header only, don't download entire file
            return requests.head(current).status_code < 400
        except Exception as e:
            console = Console()
            console.print(f'\n  [red]Invalid http request[/red]: {str(e)}')
        return False

    if current.startswith('s3://'):
        return True

    # File path
    print(current)
    return os.path.exists(current)


def get_s3_bucket_region(s3_bucket):
    try:
        s3_url = f'https://{s3_bucket}.s3.amazonaws.com/'
        response = requests.get(s3_url)
        if response.status_code == 403:
            return response.headers.get('x-amz-bucket-region')
        else:
            return ''
    except Exception:
        return ''


def get_aws_credentials():
    if os.path.exists(AWS_CREDENTIAL_PATH):
        config = configparser.RawConfigParser()
        config.read(AWS_CREDENTIAL_PATH)
        profiles = config.sections()
        if len(profiles) == 0:
            return None, None
        return config.get(profiles[0], 'aws_access_key_id'), config.get(profiles[0], 'aws_secret_access_key')

    return os.getenv('AWS_ACCESS_KEY_ID'), os.getenv('AWS_SECRET_ACCESS_KEY')


class DuckDBDataSource(DataSource):

    def __init__(self, name, **kwargs):
        super().__init__(name, 'duckdb', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            PathField('path', description='Path of database file'),
            TextField('schema', description='Name of the schema', optional=True, default='main'),
        ]

    def validate(self):
        return self._validate_required_fields()

    def verify_connector(self):
        try:
            import duckdb
            import duckdb_engine
            import chardet
        except Exception as e:
            return PipeRiderConnectorError(str(e), self.type_name)

    def to_database_url(self):
        credential = self.credential
        dbpath = credential.get('path')
        duckdb_path = os.path.abspath(dbpath)
        if not os.path.exists(duckdb_path):
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=duckdb_path)
        return f"duckdb:///{duckdb_path}"

    def _formalize_table_name(self, name):
        if not name:
            return f'{self.type_name}_table'

        if not name[0].isalpha():
            name = 'TABLE_' + name

        if len(name) > 120:
            name = name[:115] + '__'

        return name.replace('.', '_')


class CsvDataSource(DuckDBDataSource):
    def __init__(self, name, **kwargs):
        super(DuckDBDataSource, self).__init__(name, 'csv', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            PathField('path', description='Path of csv file'),
        ]

    def to_database_url(self):
        return 'duckdb:///:memory:'

    def create_engine(self):
        credential = self.credential
        csv_path = os.path.abspath(credential.get('path'))
        if not os.path.exists(csv_path):
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=csv_path)
        table_name = self._formalize_table_name(splitext(basename(csv_path))[0])
        engine = super().create_engine()
        # Load csv file as table
        sql_query = f"CREATE TABLE '{table_name}' AS SELECT * FROM read_csv_auto('{csv_path}')"
        try:
            engine.execute('SET enable_progress_bar=true;')
            engine.execute(sql_query)
            engine.execute('SET enable_progress_bar=false;')
        except Exception as e:
            if isinstance(e, DBAPIError) and e.args[0].endswith('Invalid Error: String value is not valid UTF8'):
                encoding_detection = self.detect_file_encoding(csv_path)
                raise PipeRiderDataBaseEncodingError(csv_path, 'csv', encoding_detection['encoding'], 'UTF-8')
            raise e
        return engine

    def detect_file_encoding(self, file):
        import chardet
        with open(file, "rb") as fd:
            rawdata = fd.read()
            return chardet.detect(rawdata)


class ParquetDataSource(DuckDBDataSource):

    def __init__(self, name, **kwargs):
        super(DuckDBDataSource, self).__init__(name, 'parquet', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            TextField('path',
                      description='Path of the Parquet file. (Support: file path, http and s3)',
                      validate=_parquet_path_validate_func),
        ]

    def to_database_url(self):
        return 'duckdb:///:memory:'

    def _extract_parquet_path(self):
        credential = self.credential
        parquet_path = credential.get('path')

        if _parquet_path_validate_func is False:
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=parquet_path)

        if parquet_path.startswith('https://') \
            or parquet_path.startswith('http://'):
            return 'http', parquet_path

        if parquet_path.startswith('s3://'):
            aws_access_key_id, aws_secret_access_key = get_aws_credentials()
            if aws_access_key_id is None or aws_secret_access_key is None:
                raise AwsCredentialsError('No AWS credentials found')
            s3_uri = urlparse(parquet_path)
            s3_bucket = s3_uri.netloc
            s3_region = get_s3_bucket_region(s3_bucket)
            if not s3_region:
                raise AwsUnExistedS3Bucket(s3_bucket)
            return 's3', parquet_path

        return 'file', os.path.abspath(parquet_path)

    def create_engine(self):
        type, parquet_path = self._extract_parquet_path()
        engine = super().create_engine()

        if type == 'http':
            # HTTP
            engine.execute('INSTALL httpfs')
            engine.execute('LOAD httpfs')
            url = urlparse(parquet_path)
            table_name = self._formalize_table_name(splitext(os.path.basename(url.path))[0])
            pass
        elif type == 's3':
            # S3
            s3_uri = urlparse(parquet_path)
            engine.execute('INSTALL httpfs')
            engine.execute('LOAD httpfs')
            aws_access_key_id, aws_secret_access_key = get_aws_credentials()
            engine.execute(f'set s3_access_key_id="{aws_access_key_id}"')
            engine.execute(f'set s3_secret_access_key="{aws_secret_access_key}"')
            engine.execute(f'set s3_region="{get_s3_bucket_region(s3_uri.netloc)}"')
            table_name = self._formalize_table_name(splitext(os.path.basename(s3_uri.path))[0])
        else:
            # File
            table_name = self._formalize_table_name(splitext(basename(parquet_path))[0])

        sql_query = f"CREATE TABLE '{table_name}' AS SELECT * FROM read_parquet('{parquet_path}')"
        engine.execute('SET enable_progress_bar=true;')
        engine.execute(sql_query)
        engine.execute('SET enable_progress_bar=false;')
        return engine



import json
import os

from rich.console import Console
from rich.table import Table

from piperider_cli import raise_exception_when_directory_not_writable
from piperider_cli.assertion_engine import AssertionEngine
from piperider_cli.error import PipeRiderNoProfilingResultError
from piperider_cli.filesystem import FileSystem

console = Console()


def _get_run_json_path(filesystem: FileSystem, input=None):
    console = Console()
    run_json = None
    if input:
        if not os.path.exists(input):
            console.print(f'[bold red]Error: {input} not found[/bold red]')
            return
        if os.path.isdir(input):
            run_json = os.path.join(input, 'run.json')
        else:
            run_json = input
    else:
        latest = os.path.join(filesystem.get_output_dir(), 'latest')
        run_json = os.path.join(latest, 'run.json')
    return run_json


def _validate_input_result(result):
    for f in ['tables', 'id', 'created_at', 'datasource']:
        if f not in result:
            return False
    return True


class AssertionGenerator():
    @staticmethod
    def exec(input=None, report_dir: str = None):
        filesystem = FileSystem(report_dir=report_dir)
        raise_exception_when_directory_not_writable(report_dir)

        run_json_path = _get_run_json_path(filesystem, input)
        if not os.path.isfile(run_json_path):
            raise PipeRiderNoProfilingResultError(run_json_path)

        with open(run_json_path) as f:
            profiling_result = json.loads(f.read())
        if not _validate_input_result(profiling_result):
            console.print(f'[bold red]Error: {run_json_path} is invalid[/bold red]')
            return
        console.print(f'[bold dark_orange]Generating recommended assertions from:[/bold dark_orange] {run_json_path}')

        # Generate recommended assertions
        assertion_engine = AssertionEngine(None)
        assertion_engine.load_assertions(profiler_result=profiling_result)
        assertion_exist = True if assertion_engine.assertions_content else False
        console.rule('Generating Recommended Assertions')
        recommended_assertions = assertion_engine.generate_recommended_assertions(profiling_result,
                                                                                  assertion_exist=assertion_exist)
        # Show the generated recommended assertions
        ascii_table = Table(show_header=True, header_style="bold magenta")
        ascii_table.add_column('Table Name', style="bold yellow")
        ascii_table.add_column('Column Name', style="bold blue")
        ascii_table.add_column('Test Function', style="bold green")
        ascii_table.add_column('Asserts', style="bold")

        for assertion in assertion_engine.recommender.generated_assertions:
            assert_values = str(assertion.asserts) if assertion.asserts else ''
            ascii_table.add_row(
                assertion.table,
                assertion.column,
                assertion.name,
                assert_values,
            )

        console.print(ascii_table)

        # Show the recommended assertions files
        console.rule('Generated Recommended Assertions')
        for f in recommended_assertions:
            console.print(f'[bold green]Recommended Assertion[/bold green]: {f}')
